from mage_ai.orchestration.db.models.schedules import PipelineRun
from mage_ai.orchestration.execution_process_manager import ExecutionProcessManager
from mage_ai.tests.base_test import DBTestCase
from mage_ai.tests.factory import create_pipeline_with_blocks
from unittest.mock import MagicMock, patch


class ExecutionProcessManagerTests(DBTestCase):
    @classmethod
    def setUpClass(self):
        super().setUpClass()
        self.pipeline = create_pipeline_with_blocks(
            'test pipeline',
            self.repo_path,
        )

    def test_has_pipeline_process(self):
        manager = ExecutionProcessManager()
        self.assertFalse(manager.has_pipeline_process(1))
        proc = MagicMock()
        manager.pipeline_processes[1] = proc
        self.assertTrue(manager.has_pipeline_process(1))

    def test_set_block_process(self):
        manager = ExecutionProcessManager()
        proc = MagicMock()
        manager.set_block_process(10, 50, proc)
        self.assertEqual(manager.block_processes[10][50], proc)

    def test_terminate_pipeline_process(self):
        manager = ExecutionProcessManager()
        proc = MagicMock()
        manager.set_pipeline_process(1, proc)
        with patch.object(proc, 'is_alive', return_value=True):
            with patch.object(proc, 'terminate') as mock_terminate:
                manager.terminate_pipeline_process(1)
                mock_terminate.assert_called_once()

    def test_set_pipeline_process(self):
        manager = ExecutionProcessManager()
        proc = MagicMock()
        with patch.object(manager, 'terminate_pipeline_process') as mock_terminate:
            manager.set_pipeline_process(10, proc)
            self.assertEqual(manager.pipeline_processes[10], proc)
            mock_terminate.assert_called_once_with(10)
        self.assertTrue(manager.has_pipeline_process(10))

    def test_clean_up_processes(self):
        pipeline_run1 = PipelineRun.create(pipeline_uuid='test_pipeline')
        pipeline_run2 = PipelineRun.create(pipeline_uuid='test_pipeline')
        pipeline_run3 = PipelineRun.create(pipeline_uuid='test_pipeline')

        block_runs = pipeline_run1.block_runs
        pipeline_run2.update(status=PipelineRun.PipelineRunStatus.CANCELLED)

        manager = ExecutionProcessManager()
        proc_block_terminated = MagicMock()
        proc_block_terminated.is_alive = MagicMock(return_value=False)
        proc_block_cancelled = MagicMock()
        proc_block_live = MagicMock()
        proc_pipeline_terminated = MagicMock()
        proc_pipeline_terminated.is_alive = MagicMock(return_value=False)
        proc_pipeline_cancelled = MagicMock()
        proc_pipeline_live = MagicMock()
        manager.set_block_process(pipeline_run1.id, block_runs[0].id, proc_block_terminated)
        manager.set_block_process(pipeline_run2.id, block_runs[1].id, proc_block_cancelled)
        manager.set_block_process(pipeline_run1.id, block_runs[2].id, proc_block_live)
        manager.set_pipeline_process(pipeline_run1.id, proc_pipeline_terminated)
        manager.set_pipeline_process(pipeline_run2.id, proc_pipeline_cancelled)
        manager.set_pipeline_process(pipeline_run3.id, proc_pipeline_live)

        manager.clean_up_processes(include_child_processes=False)
        self.assertEqual(
            manager.block_processes,
            {pipeline_run1.id: {block_runs[2].id: proc_block_live}},
        )
        self.assertEqual(
            manager.pipeline_processes,
            {pipeline_run3.id: proc_pipeline_live},
        )



{% extends "transformers/transformer_actions/action.jinja" %}
{% block action %}
    """
    Execute Transformer Action: ActionType.REMOVE

    Docs: https://docs.mage.ai/guides/transformer-blocks#remove-rows
    """
    action = build_transformer_action(
        df,
        action_type=ActionType.REMOVE,
        axis=Axis.ROW,
        options={'rows': []},  # Specify indices of rows to remove
    )
{% endblock %}



"""Add passed_sla to pipeline run

Revision ID: 5cd59ec4cf1d
Revises: 84de4cdd6126
Create Date: 2022-10-21 11:39:38.166335

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '5cd59ec4cf1d'
down_revision = '84de4cdd6126'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('pipeline_run', sa.Column('passed_sla', sa.Boolean(), default=False))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('pipeline_run', 'passed_sla')
    # ### end Alembic commands ###



from mage_ai.services.google_chat.config import GoogleChatConfig
import requests


def send_google_chat_message(
    config: GoogleChatConfig,
    message: str,
    title: str = 'Mage pipeline run status logs',
) -> None:
    requests.post(
        url=config.webhook_url,
        json={
            'text': '*{title}*\n{message}'.format(title=title, message=message),
        },
    )



from unittest.mock import ANY, patch

from mage_ai.orchestration.notification.config import NotificationConfig
from mage_ai.orchestration.notification.sender import NotificationSender
from mage_ai.tests.base_test import DBTestCase
from mage_ai.tests.factory import create_pipeline, create_pipeline_run_with_schedule
from mage_ai.tests.orchestration.notification.constants import (
    EMAIL_NOTIFICATION_CONFIG,
    OPSGENIE_NOTIFICATION_CONFIG,
    SLACK_NOTIFICATION_CONFIG,
    SLACK_NOTIFICATION_CONFIG_WITH_CUSTOM_TEMPLATE,
    TEAMS_NOTIFICATION_CONFIG,
    TEAMS_NOTIFICATION_CONFIG_NO_ALERT_ON,
)


class NotificationSenderTests(DBTestCase):
    @classmethod
    def setUpClass(self):
        super().setUpClass()
        self.pipeline = create_pipeline('test pipeline', self.repo_path)
        self.pipeline_run = create_pipeline_run_with_schedule(pipeline_uuid='test_pipeline')

    @patch('mage_ai.orchestration.notification.sender.send_slack_message')
    @patch('mage_ai.orchestration.notification.sender.send_email')
    def test_send_pipeline_run_success_message(self, mock_send_email, mock_send_slack):
        notification_config = NotificationConfig.load(config=EMAIL_NOTIFICATION_CONFIG)
        sender = NotificationSender(config=notification_config)
        pipeline_run = self.__class__.pipeline_run
        sender.send_pipeline_run_success_message(self.__class__.pipeline, pipeline_run)
        self.assertEqual(mock_send_slack.call_count, 0)
        email_subject = 'Successfully ran Pipeline test_pipeline'
        email_content = (
            'Successfully ran Pipeline `test_pipeline` '
            f'with Trigger {pipeline_run.pipeline_schedule.id} '
            f'`{pipeline_run.pipeline_schedule.name}` '
            f'at execution time `{pipeline_run.execution_date}`.'
        )
        email_content += (
            '\nOpen http://localhost:6789/pipelines/test_pipeline/runs/'
            f'{pipeline_run.id} to check pipeline run results and logs.'
        )
        mock_send_email.assert_called_once_with(
            notification_config.email_config,
            subject=email_subject,
            message=email_content,
        )

    @patch('mage_ai.orchestration.notification.sender.send_slack_message')
    @patch('mage_ai.orchestration.notification.sender.send_email')
    def test_send_pipeline_run_failure_message(self, mock_send_email, mock_send_slack):
        notification_config = NotificationConfig.load(config=SLACK_NOTIFICATION_CONFIG)
        sender = NotificationSender(config=notification_config)
        pipeline_run = self.__class__.pipeline_run
        sender.send_pipeline_run_failure_message(self.__class__.pipeline, pipeline_run)
        self.assertEqual(mock_send_email.call_count, 0)
        message = (
            'Failed to run Pipeline `test_pipeline` '
            f'with Trigger {pipeline_run.pipeline_schedule.id} '
            f'`{pipeline_run.pipeline_schedule.name}` '
            f'at execution time `{pipeline_run.execution_date}`.\n'
            f'Open http://localhost:6789/pipelines/test_pipeline/runs/'
            f'{pipeline_run.id} to check pipeline run results and logs.'
        )
        mock_send_slack.assert_called_once_with(
            notification_config.slack_config,
            message,
        )

    @patch('mage_ai.orchestration.notification.sender.send_slack_message')
    @patch('mage_ai.orchestration.notification.sender.send_email')
    def test_send_pipeline_run_failure_message_with_custom_template(
        self,
        mock_send_email,
        mock_send_slack,
    ):
        notification_config = NotificationConfig.load(
            config=SLACK_NOTIFICATION_CONFIG_WITH_CUSTOM_TEMPLATE)
        sender = NotificationSender(config=notification_config)
        pipeline_run = self.__class__.pipeline_run
        sender.send_pipeline_run_failure_message(self.__class__.pipeline, pipeline_run)
        self.assertEqual(mock_send_email.call_count, 0)
        message = (
            'Failed to execute pipeline http://localhost:6789/pipelines/test_pipeline/runs/'
            f'{pipeline_run.id}. Pipeline uuid: test_pipeline. '
            f'Trigger name: {pipeline_run.pipeline_schedule.name}.'
        )
        mock_send_slack.assert_called_once_with(
            notification_config.slack_config,
            message,
        )

    @patch('mage_ai.orchestration.notification.sender.send_teams_message')
    @patch('mage_ai.orchestration.notification.sender.send_email')
    def test_send_pipeline_run_failure_message_using_teams(
        self,
        mock_send_email,
        mock_send_teams_message,
    ):
        notification_config = NotificationConfig.load(config=TEAMS_NOTIFICATION_CONFIG)
        sender = NotificationSender(config=notification_config)
        pipeline_run = self.__class__.pipeline_run
        sender.send_pipeline_run_failure_message(self.__class__.pipeline, pipeline_run)
        self.assertEqual(mock_send_email.call_count, 0)
        message = (
            'Failed to run Pipeline `test_pipeline` '
            f'with Trigger {pipeline_run.pipeline_schedule.id} '
            f'`{pipeline_run.pipeline_schedule.name}` '
            f'at execution time `{pipeline_run.execution_date}`.'
        )
        mock_send_teams_message.assert_called_once_with(
            notification_config.teams_config,
            message,
        )

    @patch('mage_ai.orchestration.notification.sender.send_teams_message')
    @patch('mage_ai.orchestration.notification.sender.send_email')
    def test_alert_on_configuration(self, mock_send_email, mock_send_teams_message):
        notification_config = NotificationConfig.load(config=TEAMS_NOTIFICATION_CONFIG)
        sender = NotificationSender(config=notification_config)
        pipeline_run = self.__class__.pipeline_run
        sender.send_pipeline_run_success_message(self.__class__.pipeline, pipeline_run)
        self.assertEqual(mock_send_email.call_count, 0)
        message = (
            'Successfully ran Pipeline `test_pipeline` '
            f'with Trigger {pipeline_run.pipeline_schedule.id} '
            f'`{pipeline_run.pipeline_schedule.name}` '
            f'at execution time `{pipeline_run.execution_date}`.'
        )
        mock_send_teams_message.assert_called_once_with(
            notification_config.teams_config,
            message,
        )

    @patch('mage_ai.orchestration.notification.sender.send_teams_message')
    @patch('mage_ai.orchestration.notification.sender.send_email')
    def test_alert_on_configuration_empty(self, mock_send_email, mock_send_teams_message):
        notification_config = NotificationConfig.load(config=TEAMS_NOTIFICATION_CONFIG_NO_ALERT_ON)
        sender = NotificationSender(config=notification_config)
        pipeline_run = self.__class__.pipeline_run
        sender.send_pipeline_run_failure_message(self.__class__.pipeline, pipeline_run)
        self.assertEqual(mock_send_email.call_count, 0)
        self.assertEqual(mock_send_teams_message.call_count, 0)

    @patch('mage_ai.orchestration.notification.sender.send_opsgenie_alert')
    @patch('mage_ai.orchestration.notification.sender.send_email')
    def test_send_pipeline_run_failure_message_using_opsgenie(
            self,
            mock_send_email,
            mock_send_opsgenie_message,
    ):
        notification_config = NotificationConfig.load(config=OPSGENIE_NOTIFICATION_CONFIG)
        sender = NotificationSender(config=notification_config)
        pipeline_run = self.__class__.pipeline_run
        sender.send_pipeline_run_failure_message(self.__class__.pipeline, pipeline_run)
        self.assertEqual(mock_send_email.call_count, 0)
        mock_send_opsgenie_message.assert_called_once_with(
            notification_config.opsgenie_config,
            message=ANY,
            description=ANY,
        )



import multiprocessing
import time
import traceback
from enum import Enum

import sentry_sdk

from mage_ai.orchestration.db.database_manager import database_manager
from mage_ai.orchestration.db.process import create_process
from mage_ai.server.logger import Logger
from mage_ai.settings import SENTRY_DSN, SENTRY_TRACES_SAMPLE_RATE

SCHEDULER_AUTO_RESTART_INTERVAL = 20_000    # in milliseconds

logger = Logger().new_server_logger(__name__)


def run_scheduler():
    from mage_ai.orchestration.triggers.loop_time_trigger import LoopTimeTrigger

    sentry_dsn = SENTRY_DSN
    if sentry_dsn:
        sentry_sdk.init(
            sentry_dsn,
            traces_sample_rate=SENTRY_TRACES_SAMPLE_RATE,
        )

    try:
        database_manager.run_migrations()
    except Exception:
        traceback.print_exc()
    try:
        LoopTimeTrigger().start()
    except Exception as e:
        traceback.print_exc()
        raise e


class SchedulerManager:
    """
    Singleton class to manage scheduler process.
    """

    class SchedulerStatus(str, Enum):
        RUNNING = 'running'
        STOPPED = 'stopped'

    def __init__(self):
        self.scheduler_process: multiprocessing.Process = None
        self.status = self.SchedulerStatus.STOPPED

    @property
    def is_alive(self):
        return self.scheduler_process is not None and self.scheduler_process.is_alive()

    def get_status(self, auto_restart: bool = False):
        if auto_restart and self.status == self.SchedulerStatus.RUNNING and not self.is_alive:
            logger.info('Restarting pipeline scheduler.')
            self.start_scheduler()
        if self.is_alive:
            return SchedulerManager.SchedulerStatus.RUNNING
        return SchedulerManager.SchedulerStatus.STOPPED

    def start_scheduler(self, foreground: bool = False):
        logger.info('Start scheduler.')
        if self.is_alive:
            return

        proc = create_process(target=run_scheduler)
        proc.start()
        self.scheduler_process = proc
        self.status = self.SchedulerStatus.RUNNING
        if foreground:
            while True:
                check_scheduler_status()
                time.sleep(SCHEDULER_AUTO_RESTART_INTERVAL / 1000)

    def stop_scheduler(self):
        logger.info('Stop scheduler.')
        if self.is_alive:
            self.scheduler_process.terminate()
            self.scheduler_process = None
            self.status = self.SchedulerStatus.STOPPED


scheduler_manager = SchedulerManager()


def check_scheduler_status():
    status = scheduler_manager.get_status(auto_restart=True)
    logger.info(f'Scheduler status: {status}.')



from mage_ai.api.presenters.BasePresenter import BasePresenter


class OutputPresenter(BasePresenter):
    default_attributes = [
        'sample_data',
        'shape',
        'text_data',
        'type',
        'variable_uuid',
    ]



"""Add started_at to block run

Revision ID: b01be687e537
Revises: 067326f43bc3
Create Date: 2022-09-07 10:37:29.955476

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'b01be687e537'
down_revision = '067326f43bc3'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('block_run', sa.Column('started_at', sa.DateTime(timezone=True), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('block_run', 'started_at')
    # ### end Alembic commands ###



from mage_ai.orchestration.triggers.time_trigger import TimeTrigger
import time


class LoopTimeTrigger(TimeTrigger):
    def start(self) -> None:
        while True:
            self.last_run_time = int(time.time())
            self.run()
            current_time = int(time.time())
            sleep_time = self.trigger_interval - (current_time - self.last_run_time)
            if sleep_time > 0:
                time.sleep(sleep_time)



from mage_ai.services.aws.s3.s3 import Client
from mage_ai.tests.base_test import TestCase
from unittest.mock import MagicMock, patch


class ClientTests(TestCase):
    @patch('mage_ai.services.aws.s3.s3.boto3')
    @patch('mage_ai.services.aws.s3.s3.botocore')
    @patch('mage_ai.services.aws.s3.s3.s3transfer')
    def test_listdir(self, mock_s3transfer, mock_botocore, mock_boto3):
        mock_client = MagicMock()
        mock_client.list_objects_v2 = MagicMock(return_value={
            'Contents': [
                {'Key': 'test_prefix/file1'},
                {'Key': 'test_prefix/file2.csv'},
                {'Key': 'test_prefix/file3.parquet'},
                {'Key': 'test_prefix/file4.csv'},
            ],
            'CommonPrefixes': [{'Prefix': 'test_prefix/test_folder/'}],
        })
        s3_client = Client('test_bucket', client=mock_client)

        results = s3_client.listdir('test_prefix', suffix='.csv')
        mock_client.list_objects_v2.assert_called_once_with(
            Bucket='test_bucket',
            MaxKeys=10000,
            Prefix='test_prefix',
            Delimiter='/',
        )
        self.assertEqual(results, [
            'test_prefix/file2.csv',
            'test_prefix/file4.csv',
            'test_prefix/test_folder/'
        ])

    @patch('mage_ai.services.aws.s3.s3.boto3')
    @patch('mage_ai.services.aws.s3.s3.botocore')
    @patch('mage_ai.services.aws.s3.s3.s3transfer')
    def test_list_objects(self, mock_s3transfer, mock_botocore, mock_boto3):
        mock_client = MagicMock()
        mock_client.list_objects_v2 = MagicMock(return_value={
            'Contents': [
                {'Key': 'test_prefix/file1'},
                {'Key': 'test_prefix/file2.csv'},
                {'Key': 'test_prefix/file3.parquet'},
                {'Key': 'test_prefix/test_folder/file4.csv'},
            ],
        })
        s3_client = Client('test_bucket', client=mock_client)

        results = s3_client.list_objects('test_prefix', suffix='.csv')
        mock_client.list_objects_v2.assert_called_once_with(
            Bucket='test_bucket',
            MaxKeys=10000,
            Prefix='test_prefix',
        )
        self.assertEqual(results, [
            'test_prefix/file2.csv',
            'test_prefix/test_folder/file4.csv',
        ])



from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.orchestration.db import safe_db_query
from mage_ai.orchestration.db.models.oauth import Role


class RoleResource(GenericResource):
    model_class = Role

    @classmethod
    @safe_db_query
    def collection(self, query, meta, user, **kwargs):
        from mage_ai.orchestration.db.models.oauth import Permission

        limit_roles = query.get('limit_roles', [None])
        if limit_roles:
            limit_roles = limit_roles[0]

        entity = query.get('entity', [None])
        if entity:
            entity = entity[0]

        entity_ids = query.get('entity_ids[]', [])
        if entity_ids:
            entity_ids = entity_ids[0]
        if entity_ids:
            entity_ids = entity_ids.split(',')

        permissions_query = Permission.query
        if entity:
            permissions_query = permissions_query.filter(
                Permission.entity == entity,
            )
            if entity != Permission.Entity.GLOBAL and entity_ids:
                permissions_query = permissions_query.filter(
                    Permission.entity_id.in_(entity_ids),
                )
            permissions = permissions_query.all()
            roles = []
            for permission in permissions:
                roles.append(permission.role)
        else:
            roles = Role.query.all()
        access = user.get_access(Permission.Entity.PROJECT, get_repo_path())

        if (access & Permission.Access.OWNER == 0) and limit_roles:
            role_access = Permission.Access.EDITOR | Permission.Access.VIEWER
            roles = list(filter(
                lambda role: role.get_access(
                    Permission.Entity.PROJECT,
                    get_repo_path(),
                ) | role_access == role_access,  # Only editors and viewers
                roles,
            ))

        return self.build_result_set(roles, user, **kwargs)



from jinja2 import Template
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.data_preparation.shared.utils import get_template_vars
from mage_ai.data_integrations.utils.parsers import NoDatesSafeLoader
from mage_ai.shared.dates import n_days_ago
from mage_ai.shared.hash import merge_dict
from mage_ai.shared.parsers import encode_complex
from typing import Dict, List
import json
import simplejson
import yaml

KEY_PATTERNS = '_patterns'
PATTERN_KEY_DESTINATION_TABLE = 'destination_table'


def get_settings(block, variables: Dict = {}, pipeline: 'Pipeline' = None) -> Dict:
    return __get_settings(block.file_path, variables, pipeline=pipeline)


def __get_settings(absolute_file_path, variables: Dict = {}, pipeline: 'Pipeline' = None) -> Dict:
    settings = interpolate_variables_for_block_settings(absolute_file_path, variables)
    settings_raw = load_yaml_file(absolute_file_path)

    config = settings_raw['config']
    patterns = config.get(KEY_PATTERNS, {})
    destination_table_pattern = patterns.get(PATTERN_KEY_DESTINATION_TABLE)

    if not settings.get('catalog') \
        and pipeline \
        and pipeline.data_integration \
            and 'catalog' in pipeline.data_integration:
        settings['catalog'] = interpolate_variables(
            json.dumps(pipeline.data_integration['catalog']),
            variables,
        )

    if destination_table_pattern:
        for stream in settings['catalog']['streams']:
            tap_stream_id = stream['tap_stream_id']
            destination_table_init = stream.get(PATTERN_KEY_DESTINATION_TABLE)

            stream[PATTERN_KEY_DESTINATION_TABLE] = interpolate_string(
                destination_table_pattern,
                merge_dict(
                    variables,
                    {
                        'stream': tap_stream_id,
                        PATTERN_KEY_DESTINATION_TABLE: destination_table_init,
                    },
                ),
            )

    return settings


def get_catalog(block, variables: Dict = {}, pipeline: 'Pipeline' = None) -> Dict:
    return get_settings(block, variables, pipeline=pipeline)['catalog']


def get_catalog_by_stream(
    absolute_file_path,
    stream_id: str,
    variables: Dict = {},
    pipeline: 'Pipeline' = None,
) -> Dict:
    catalog = __get_settings(absolute_file_path, variables, pipeline=pipeline)['catalog']
    for stream in catalog['streams']:
        tap_stream_id = stream['tap_stream_id']
        if tap_stream_id == stream_id:
            return stream
    return None


def build_catalog_json(
    absolute_file_path: str,
    variables: Dict,
    selected_streams: List[str] = None,
    pipeline: 'Pipeline' = None,
) -> str:
    streams = []

    settings = __get_settings(absolute_file_path, variables, pipeline=pipeline)
    if 'catalog' in settings and 'streams' in settings['catalog']:
        for stream in settings['catalog']['streams']:
            tap_stream_id = stream['tap_stream_id']
            if not selected_streams or tap_stream_id in selected_streams:
                streams.append(stream)

    catalog = dict(streams=streams)

    return simplejson.dumps(
        catalog,
        default=encode_complex,
        ignore_nan=True,
    )


def build_config_json(
    absolute_file_path: str,
    variables: Dict,
    override: Dict = None,
) -> str:
    config = interpolate_variables_for_block_settings(
        absolute_file_path,
        variables,
    )['config']

    if override:
        config.update(override)

    return simplejson.dumps(
        config,
        default=encode_complex,
        ignore_nan=True,
    )


def load_yaml(text: str):
    return yaml.load(text, Loader=NoDatesSafeLoader)


def load_yaml_file(absolute_file_path: str) -> Dict:
    with open(absolute_file_path, 'r') as f:
        return load_yaml(f.read())


def interpolate_variables_for_block_settings(
    absolute_file_path: str,
    variables: Dict,
) -> Dict:
    with open(absolute_file_path, 'r') as f:
        return interpolate_variables(f.read(), variables)


def interpolate_string(text: str, variables: Dict) -> str:
    variables = dict() if variables is None else variables
    kwargs = dict(
        variables=lambda x: variables.get(x),
        n_days_ago=n_days_ago,
        **get_template_vars(),
    )

    return Template(text).render(**kwargs)


def interpolate_variables(
    text: str,
    variables: Dict,
) -> Dict:
    settings_string = interpolate_string(text, variables)

    return load_yaml(settings_string)



from datetime import datetime
from mage_ai.shared.hash import merge_dict
from typing import Dict
import json
import re


def print_log_from_line(
    line: str,
    logger=None,
    logging_tags: Dict = dict(),
    tags: Dict = {},
):
    from mage_integrations.utils.logger.constants import (
        LOG_LEVEL_ERROR,
        LOG_LEVEL_EXCEPTION,
        TYPE_LOG,
    )

    try:
        data = json.loads(line)

        if type(data) is not dict:
            if type(data) is list and len(data) >= 1:
                print(json.dumps(data))
            return

        message = data.get('message')
        tags1 = data.get('tags')

        if 'timestamp' in data:
            message = datetime.fromtimestamp(data['timestamp']).strftime('%Y-%m-%dT%H:%M:%S') \
                        + ' ' + str(message)

        if message and (
            re.match(
                '^[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}:[0-9]{2} Unable to parse:',
                message,
            ) or
            re.match('Unable to parse:', message)
        ):
            return

        if TYPE_LOG == data.get('type'):
            if logger:
                data2 = data.copy()
                if 'message' in data2:
                    del data2['message']
                if 'tags' in data2:
                    del data2['tags']

                updated_tags = tags1
                try:
                    updated_tags.update(data2)
                    logger.info(
                        message,
                        **merge_dict(
                            logging_tags,
                            dict(tags=merge_dict(tags, updated_tags)),
                        )
                    )
                except Exception:
                    pass
            else:
                print(json.dumps(data))
        if data.get('level') in [LOG_LEVEL_ERROR, LOG_LEVEL_EXCEPTION]:
            if message:
                if tags1:
                    message = f'{message} {json.dumps(tags1)}'
            else:
                message = 'Exception raised, please check logs for more details.'

            raise Exception(message)
    except json.decoder.JSONDecodeError:
        pass


def print_logs_from_output(
    output: str,
    logger=None,
    logging_tags: Dict = dict(),
    tags: Dict = {},
):
    for line in output.split('\n'):
        print_log_from_line(
            line,
            logger=logger,
            logging_tags=logging_tags,
            tags=tags,
        )



from mage_ai.data_cleaner.analysis import charts
from mage_ai.data_cleaner.analysis.constants import (
    DATA_KEY_CHARTS,
    DATA_KEY_CORRELATION,
    DATA_KEY_TIME_SERIES,
)
from mage_ai.data_cleaner.column_types.constants import ColumnType
from mage_ai.data_cleaner.shared.utils import clean_dataframe, is_numeric_dtype
from mage_ai.shared.logger import timer, VerboseFunctionExec
from mage_ai.shared.hash import merge_dict
import logging

DD_KEY = 'lambda.analysis_calculator'
TIMESERIES_COLUMN_TYPES = frozenset(
    [
        ColumnType.CATEGORY,
        ColumnType.CATEGORY_HIGH_CARDINALITY,
        ColumnType.NUMBER,
        ColumnType.NUMBER_WITH_DECIMALS,
        ColumnType.TRUE_OR_FALSE,
    ]
)
VERBOSE = False

logger = logging.getLogger(__name__)


def increment(metric, tags={}):
    pass


class AnalysisCalculator:
    def __init__(
        self,
        df,
        column_types,
        statistics,
        verbose=False,
        **kwargs,
    ):
        self.df = df
        self.column_types = column_types
        self.features = [{'uuid': col, 'column_type': column_types.get(col)} for col in df.columns]
        self.verbose = verbose

    def process(self, df, is_clean=False):
        with VerboseFunctionExec('Generating visualizations', verbose=self.verbose):
            return self.__process(df, is_clean=is_clean)

    def __process(self, df, is_clean=False):
        increment(f'{DD_KEY}.process.start', self.tags)
        df_columns = df.columns

        features_to_use = self.features
        datetime_features_to_use = [f for f in self.datetime_features if f['uuid'] in df_columns]
        numeric_features_to_use = self.numeric_features

        if not is_clean:
            df_clean = clean_dataframe(df, self.column_types, dropna=False)
        else:
            df_clean = df

        arr_args_1 = ([df_clean for _ in features_to_use],)
        arr_args_2 = (features_to_use,)

        data_for_columns = [d for d in map(self.calculate_column, *arr_args_1, *arr_args_2)]

        correlation_data = self.calculate_correlation_data(df)
        time_series_charts = self.calculate_timeseries_data(df)
        for d in data_for_columns:
            fuuid = d['feature']['uuid']
            if fuuid in correlation_data:
                d[DATA_KEY_CORRELATION] = correlation_data[fuuid]
            if fuuid in time_series_charts:
                d[DATA_KEY_TIME_SERIES] = time_series_charts[fuuid]

        overview = charts.build_overview_data(
            df,
            datetime_features_to_use,
            numeric_features_to_use,
        )

        correlation_overview = []
        for d in data_for_columns:
            corr = d.get(DATA_KEY_CORRELATION)
            if corr:
                correlation_overview.append(
                    {
                        'feature': d['feature'],
                        DATA_KEY_CORRELATION: corr,
                    }
                )

        increment(f'{DD_KEY}.process.succeeded', self.tags)

        return data_for_columns, merge_dict(
            overview,
            {
                DATA_KEY_CORRELATION: correlation_overview,
            },
        )

    @property
    def features_by_uuid(self):
        data = {}
        for feature in self.features:
            data[feature['uuid']] = feature

        return data

    @property
    def datetime_features(self):
        return [f for f in self.features if f['column_type'] == ColumnType.DATETIME]

    @property
    def numeric_features(self):
        return [
            f['uuid']
            for f in self.features
            if is_numeric_dtype(self.df, f['uuid'], f['column_type'])
        ]

    @property
    def tags(self):
        return dict()

    def calculate_column(self, df, feature):
        with timer('analysis.calculate_column', dict(feature=feature), verbose=VERBOSE):
            try:
                return self.calculate_column_internal(df, feature)
            except Exception:
                logger.exception(f'Failed to calculate stats for column {feature}')
                return {
                    'feature': feature,
                    DATA_KEY_CHARTS: [],
                    DATA_KEY_CORRELATION: [],
                    DATA_KEY_TIME_SERIES: [],
                }

    def calculate_column_internal(self, df, feature):
        col = feature['uuid']
        column_type = feature['column_type']

        chart_data = []
        correlation = []

        is_numeric_col = is_numeric_dtype(df, col, column_type)
        if is_numeric_col:
            with timer(
                'analysis.calculate_column.build_histogram_data',
                dict(feature=feature),
                verbose=VERBOSE,
            ):
                series_cleaned = df[col].dropna()
                histogram_data = charts.build_histogram_data(col, series_cleaned, column_type)
                if histogram_data:
                    chart_data.append(histogram_data)

        return {
            'feature': feature,
            DATA_KEY_CHARTS: chart_data,
            DATA_KEY_CORRELATION: correlation,
        }

    def calculate_correlation_data(self, df):
        return charts.build_correlation_data(df)

    def calculate_timeseries_data(self, df):
        timeseries_features = [
            f for f in self.features if f['column_type'] in TIMESERIES_COLUMN_TYPES
        ]
        datetime_features_to_use = self.datetime_features

        charts_by_column = dict()
        for f in datetime_features_to_use:
            time_series_charts = charts.build_time_series_data(df, timeseries_features, f['uuid'])
            if time_series_charts is None:
                continue
            for f, chart in time_series_charts.items():
                if f not in charts_by_column:
                    charts_by_column[f] = [chart]
                else:
                    charts_by_column[f].append(chart)
        return charts_by_column



EXTENSION_DESCRIPTION = 'Create data quality checks for your data products.'
EXTENSION_NAME = 'Great Expectations'
EXTENSION_UUID = 'great_expectations'

EXTENSION_TEMPLATES = [
    dict(
        description='Base template with no additional code.',
        name='Empty template',
        path='extensions/great_expectations/default.jinja',
        uuid='default',
    ),
]



SUCCESS = 'success'
WARNING = 'warning'
CANCELLED = 'cancelled'
FAILED = 'failed'
PENDING = 'pending'
WARNING = 'warning'
SUCCESS = 'success'
QUERYING = 'querying'
PROCESSING = 'processing'
ABORTED = 'aborted'
QUEUED = 'queued'
INTERRUPTED = 'interrupted'
REPORTING = 'reporting'

TERMINAL_STATUSES = [CANCELLED, FAILED, SUCCESS, WARNING, INTERRUPTED]
PENDING_STATUSES = [
    QUEUED,
    QUERYING,
    PROCESSING,
    REPORTING,
    PENDING,
]

DEFAULT_POLL_INTERVAL = 3

HIGHTOUCH_BASE_URL = 'https://api.hightouch.com/api/v1'



from mage_ai import settings
from mage_ai.api.errors import ApiError
from mage_ai.api.resources.Resource import Resource
from mage_ai.api.resources.shared import collective_loaders
from mage_ai.api.result_set import ResultSet
from mage_ai.orchestration.db.errors import DoesNotExistError
from mage_ai.shared.hash import merge_dict
import importlib
import inflection
import inspect


class BaseResource(Resource):
    collective_loader_attr = {}
    datetime_keys = []
    model_class = None
    parent_models_attr = {}
    parent_resource_attr = {}

    @classmethod
    def policy_class(self):
        model_name = self.__name__.replace('Resource', '')
        return getattr(
            importlib.import_module(
                'mage_ai.api.policies.{}Policy'.format(model_name)),
            '{}Policy'.format(model_name),
        )

    @classmethod
    def presenter_class(self):
        model_name = self.__name__.replace('Resource', '')
        return getattr(
            importlib.import_module(
                'mage_ai.api.presenters.{}Presenter'.format(model_name)),
            '{}Presenter'.format(model_name),
        )

    @classmethod
    def collective_loader(self):
        if not self.collective_loader_attr.get(self.__name__):
            self.collective_loader_attr[self.__name__] = {}
        return self.collective_loader_attr[self.__name__]

    @classmethod
    def parent_models(self):
        if not self.parent_models_attr.get(self.__name__):
            self.parent_models_attr[self.__name__] = {}
        return self.parent_models_attr[self.__name__]

    @classmethod
    def parent_resource(self):
        if not self.parent_resource_attr.get(self.__name__):
            self.parent_resource_attr[self.__name__] = {}
        return self.parent_resource_attr[self.__name__]

    @classmethod
    def register_collective_loader(self, key, **kwargs):
        self.collective_loader()[key] = kwargs

    @classmethod
    def register_collective_loader_find(self, resource_class, **kwargs):
        attribute = kwargs.get('attribute',
                               resource_class.resource_name_singular())
        self.register_collective_loader(
            attribute,
            load=collective_loaders.build_load(
                resource_class,
                attribute=attribute),
            select=collective_loaders.build_select_find(
                '{}_id'.format(attribute)),
        )

    @classmethod
    def register_collective_loader_select(self, resource_class, **kwargs):
        attribute = kwargs.get('attribute', resource_class.resource_name())
        self.register_collective_loader(
            attribute,
            load=collective_loaders.build_load_select(
                self,
                resource_class,
                attribute=self.resource_name_singular()),
            select=collective_loaders.build_select_filter(
                '{}_id'.format(
                    self.resource_name_singular())),
        )

    @classmethod
    def register_parent_model(self, key, value):
        self.parent_models()[key] = value

    @classmethod
    def register_parent_models(self, key_values):
        for key, value in key_values.items():
            self.register_parent_model(key, value)

    @classmethod
    def register_parent_resource(self, resource_class, **kwargs):
        column_name = kwargs.get(
            'column_name',
            '{}_id'.format(resource_class.resource_name_singular()),
        )
        self.parent_resource()[column_name] = resource_class

    @classmethod
    def build_result_set(self, arr, user, **kwargs):
        return ResultSet([mod if issubclass(mod.__class__, BaseResource) else self(
            mod, user, **kwargs) for mod in arr], )

    @classmethod
    def collection(self, query, meta, user, **kwargs):
        """
        Subclasses override this method
        """
        pass

    @classmethod
    def create(self, payload, user, **kwargs):
        """
        Subclasses override this method
        """
        pass

    @classmethod
    def member(self, pk, user, **kwargs):
        """
        Subclasses override this method
        """
        pass

    @classmethod
    def before_create(self, payload, user, **kwargs):
        pass

    @classmethod
    async def process_create(self, payload, user, **kwargs):
        self.on_create_callback = None
        self.on_create_failure_callback = None
        before_create = self.before_create(payload, user, **kwargs)

        try:
            res = self.create(payload, user, **merge_dict(kwargs, {
                'before_create': before_create,
            }))
            if res and inspect.isawaitable(res):
                res = await res

            if self.on_create_callback:
                callback = self.on_create_callback(resource=res)
                if callback and inspect.isawaitable(callback):
                    await callback

            return res
        except Exception as err:
            if self.on_create_failure_callback:
                self.on_create_failure_callback(resource=res)

            raise err

    @classmethod
    async def process_collection(self, query, meta, user, **kwargs):
        res = self.collection(query, meta, user, **kwargs)
        if res and inspect.isawaitable(res):
            res = await res
        return res

    @classmethod
    async def process_member(self, pk, user, **kwargs):
        try:
            res = self.member(pk, user, **kwargs)
            if res and inspect.isawaitable(res):
                res = await res

            return res
        except DoesNotExistError as err:
            if settings.DEBUG:
                raise err
            else:
                error = ApiError.RESOURCE_NOT_FOUND
                raise ApiError(error)

    @classmethod
    def resource_name(self):
        return inflection.pluralize(self.resource_name_singular())

    @classmethod
    def resource_name_singular(self):
        return inflection.underscore(
            self.__name__.replace(
                'Resource', '')).lower()

    @classmethod
    async def get_model(self, pk):
        if self.model_class:
            return self.model_class.query.get(pk)

    def delete(self, **kwargs):
        """
        Subclasses override this method
        """
        pass

    def parent_model(self):
        return self.model_options.get('parent_model')

    async def process_delete(self, **kwargs):
        self.on_delete_callback = None
        self.on_delete_failure_callback = None

        try:
            res = self.delete(**kwargs)
            if res and inspect.isawaitable(res):
                res = await res

            if self.on_delete_callback:
                self.on_delete_callback(resource=res)

            return res
        except Exception as err:
            if self.on_delete_failure_callback:
                self.on_delete_failure_callback(resource=res)

            raise err

    async def process_update(self, payload, **kwargs):
        self.on_update_callback = None
        self.on_update_failure_callback = None

        try:
            res = self.update(payload, **kwargs)
            if res and inspect.isawaitable(res):
                res = await res

            if self.on_update_callback:
                self.on_update_callback(resource=res)

            return res
        except Exception as err:
            if self.on_update_failure_callback:
                self.on_update_failure_callback(resource=res)

            raise err

    def result_set(self):
        if self.__result_sets().get(self.__class__.__name__, None):
            return self.__result_sets()[self.__class__.__name__]
        elif not self.result_set_attr:
            self.result_set_attr = ResultSet([self])
        return self.result_set_attr

    def update(self, payload, **kwargs):
        """
        Subclasses override this method
        """
        pass

    def collective_load_for_attribute(self, key):
        k_name = self.__class__.__name__
        if self.result_set().context and self.result_set().context.data:
            loaded = self.result_set().context.data.get(k_name, {}).get(key, None)
        else:
            loaded = None
        loader = self.__class__.collective_loader().get(key, None)
        if not loaded and loader:
            loaded = loader['load'](self)
            if loaded and not isinstance(
                    loaded,
                    ResultSet) and not isinstance(
                    loaded,
                    dict):
                loaded = ResultSet(loaded)
            if not self.result_set().context.data.get(k_name):
                self.result_set().context.data[k_name] = {}
            self.result_set().context.data[k_name][key] = loaded
        return loaded

    def __result_sets(self):
        return self.model_options.get('result_sets', {})

    def __getattr__(self, name):
        def _missing(*args, **kwargs):
            loader = self.__class__.collective_loader().get(name, None)
            if loader:
                arr = self.collective_load_for_attribute(name)
                if loader['select']:
                    val = loader['select'](self, arr)
                else:
                    val = arr
            else:
                val = getattr(self.model, name)

            # This turns functions into attributes
            # if callable(val):
            #     breakpoint()
            #     return val(*args, **kwargs)
            # else:
            #     return val

            return val
        return _missing()



from mage_ai.data_preparation.models.block import Block
from mage_ai.data_preparation.models.constants import BlockType
from mage_ai.data_preparation.models.variable import (
    DATAFRAME_CSV_FILE,
    VariableType,
)
from typing import Dict, List
import jinja2
import os
import subprocess
import uuid
import pandas as pd

BLOCK_TYPE_TO_EXECUTION_TEMPLATE = {
    BlockType.DATA_LOADER: 'data_loader.jinja',
    BlockType.TRANSFORMER: 'transformer.jinja',
    BlockType.DATA_EXPORTER: 'data_exporter.jinja',
}

template_env = jinja2.Environment(
    loader=jinja2.FileSystemLoader(os.path.join(
        os.path.dirname(__file__),
        'templates',
    )),
    lstrip_blocks=True,
    trim_blocks=True,
)


def execute_r_code(
    block,
    code: str,
    execution_partition: str = None,
    global_vars: Dict = None,
):
    input_variable_objects = block.input_variable_objects(
        execution_partition=execution_partition,
    ) or []

    # Render R script with user code
    execution_code = __render_r_script(
        block,
        code,
        execution_partition=execution_partition,
        global_vars=global_vars,
        input_variable_objects=input_variable_objects,
    )
    file_path = f'/tmp/{str(uuid.uuid4())}.r'
    with open(file_path, 'w') as foutput:
        foutput.write(execution_code)

    # Convert input variable to csv format
    __convert_inputs_to_csvs(input_variable_objects)

    # Execute R script
    __execute_r_code(file_path)
    os.remove(file_path)

    output_variable_objects = block.output_variable_objects(
        execution_partition=execution_partition,
    )
    output_variable_objects = [v for v in output_variable_objects
                               if os.path.exists(os.path.join(
                                    output_variable_objects[0].variable_path,
                                    DATAFRAME_CSV_FILE,
                                ))]

    if len(output_variable_objects) > 0:
        df = pd.read_csv(os.path.join(output_variable_objects[0].variable_path, DATAFRAME_CSV_FILE))
    else:
        df = None
    return df


def __convert_inputs_to_csvs(input_variable_objects):
    for v in input_variable_objects:
        if v.variable_type == VariableType.DATAFRAME:
            v.convert_parquet_to_csv()


def __render_global_vars(global_vars: Dict = None):
    if not global_vars:
        return ''

    def format_value(val):
        if type(val) is int or type(val) is float:
            return val
        else:
            return f"'{val}'"

    var_list = [f'{k}={format_value(v)}' for k, v in global_vars.items()]
    val_list_str = ', '.join(var_list)
    return f'global_vars = c({val_list_str})'


def __render_r_script(
    block,
    code: str,
    execution_partition: str = None,
    global_vars: Dict = None,
    input_variable_objects: List = [],
):
    if block.type not in BLOCK_TYPE_TO_EXECUTION_TEMPLATE:
        raise Exception(
            f'Block execution for {block.type} with R language is not supported.',
        )
    template = template_env.get_template(BLOCK_TYPE_TO_EXECUTION_TEMPLATE[block.type])

    output_variable_object = block.variable_object(
        'output_0',
        execution_partition=execution_partition,
    )
    os.makedirs(output_variable_object.variable_path, exist_ok=True)
    output_path = os.path.join(output_variable_object.variable_path, DATAFRAME_CSV_FILE)

    global_vars_str = __render_global_vars(global_vars=global_vars)

    return template.render(
        code=code,
        global_vars=global_vars_str,
        input_paths=[os.path.join(v.variable_path, DATAFRAME_CSV_FILE)
                     for v in input_variable_objects],
        input_vars_str=', '.join([f'df_{i + 1}' for i in range(len(input_variable_objects))]),
        output_path=output_path,
    ) + '\n'


def __execute_r_code(file_path: str):
    subprocess.run(
        [
            'Rscript',
            '--vanilla',
            file_path
        ],
        check=True,
    )


class RBlock(Block):
    def _execute_block(
        self,
        output_from_input_vars,
        custom_code: str = None,
        execution_partition: str = None,
        global_vars: Dict = None,
        **kwargs,
    ) -> List:
        outputs = execute_r_code(
            self,
            custom_code or self.content,
            execution_partition=execution_partition,
            global_vars=global_vars,
        )

        if outputs is None:
            outputs = []
        if type(outputs) is not list:
            outputs = [outputs]

        return outputs



from mage_ai.cluster_manager.cluster_manager import ClusterManager
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.services.aws.emr.emr import describe_cluster, list_clusters
from mage_ai.services.aws.emr.launcher import create_cluster
from mage_ai.shared.array import find
from mage_ai.shared.hash import merge_dict
from pathlib import Path
import json
import os

CLUSTER_NAME = 'mage-data-prep'


class EmrClusterManager(ClusterManager):
    def __init__(self):
        self.active_cluster_id = None

    def list_clusters(self):
        clusters = list_clusters()['Clusters']
        valid_clusters = []
        for c in clusters:
            cluster_info = describe_cluster(c['Id'])
            cluster_tags = cluster_info['Tags']
            if find(
                lambda t: t['Key'] == 'name' and t['Value'] == CLUSTER_NAME,
                cluster_tags,
            ):
                valid_clusters.append(c)
        return [dict(
            id=c['Id'],
            name=c['Name'],
            status=c['Status']['State'],
            is_active=c['Id'] == self.active_cluster_id,
        ) for c in valid_clusters]

    def create_cluster(self):
        return create_cluster(
            get_repo_path(),
            done_status=None,
            tags=dict(name=CLUSTER_NAME),
        )

    def set_active_cluster(self, auto_selection=False, cluster_id=None):
        if cluster_id is None and auto_selection:
            clusters = self.list_clusters()
            if len(clusters) > 0:
                cluster_id = clusters[0]['id']
            else:
                self.create_cluster()
        if cluster_id is None:
            return

        self.active_cluster_id = cluster_id

        # Fetch cluster master instance public DNS
        cluster_info = describe_cluster(cluster_id)
        emr_dns = cluster_info['MasterPublicDnsName']

        # Get cluster information and update cluster url in sparkmagic config
        home_dir = str(Path.home())
        sparkmagic_config_path = os.path.join(home_dir, '.sparkmagic/config.json')
        with open(sparkmagic_config_path) as f:
            fcontent = f.read()

        config = json.loads(fcontent)
        for k, v in config.items():
            if type(v) is dict and 'url' in v:
                v['url'] = f'http://{emr_dns}:8998'

        with open(sparkmagic_config_path, 'w') as f:
            f.write(json.dumps(config))

        return merge_dict(cluster_info, dict(cluster_id=cluster_id))


emr_cluster_manager = EmrClusterManager()



from mage_ai.api.views import execute_operation
from mage_ai.server.api.base import BaseApiHandler


class ApiChildDetailHandler(BaseApiHandler):
    async def delete(self, resource, pk, child, child_pk):
        return await execute_operation(self, resource, **dict(
            pk=pk,
            child=child,
            child_pk=child_pk,
        ))

    async def get(self, resource, pk, child, child_pk):
        return await execute_operation(self, resource, **dict(
            pk=pk,
            child=child,
            child_pk=child_pk,
        ))

    async def put(self, resource, pk, child, child_pk):
        return await execute_operation(self, resource, **dict(
            pk=pk,
            child=child,
            child_pk=child_pk,
        ))


class ApiChildListHandler(BaseApiHandler):
    async def get(self, resource, pk, child):
        return await execute_operation(self, resource, **dict(
            pk=pk,
            child=child,
        ))

    async def post(self, resource, pk, child):
        return await execute_operation(self, resource, **dict(
            pk=pk,
            child=child,
        ))


class ApiResourceDetailHandler(BaseApiHandler):
    async def delete(self, resource, pk):
        return await execute_operation(self, resource, **dict(
            pk=pk,
        ))

    async def get(self, resource, pk):
        return await execute_operation(self, resource, **dict(
            pk=pk,
        ))

    async def put(self, resource, pk):
        return await execute_operation(self, resource, **dict(
            pk=pk,
        ))


class ApiResourceListHandler(BaseApiHandler):
    async def get(self, resource):
        return await execute_operation(self, resource)

    async def post(self, resource):
        return await execute_operation(self, resource)



from mage_ai.data_cleaner.transformer_actions.action_code import query_with_action_code
from mage_ai.data_cleaner.transformer_actions.udf.base import BaseUDF


class IfElse(BaseUDF):
    def execute(self):
        df_copy = self.df.copy()
        true_index = query_with_action_code(df_copy, self.code, self.kwargs).index
        arg1_type = self.options.get('arg1_type', 'value')
        arg2_type = self.options.get('arg2_type', 'value')
        arg1 = self.arguments[0]
        if arg1_type == 'column':
            arg1 = df_copy[arg1]
        arg2 = self.arguments[1]
        if arg2_type == 'column':
            arg2 = df_copy[arg2]
        df_copy.loc[true_index, 'result'] = arg1
        df_copy['result'] = df_copy['result'].fillna(arg2)
        return df_copy['result']



from mage_ai.api.oauth_scope import OauthScope
from mage_ai.api.operations import constants
from mage_ai.api.policies.BasePolicy import BasePolicy
from mage_ai.api.presenters.EventRulePresenter import EventRulePresenter


class EventRulePolicy(BasePolicy):
    @property
    def entity(self):
        return None, None


EventRulePolicy.allow_actions([
    constants.DETAIL,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_viewer_role())

EventRulePolicy.allow_read(EventRulePresenter.default_attributes + [], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.DETAIL,
], condition=lambda policy: policy.has_at_least_viewer_role())



from mage_ai.api.presenters.BasePresenter import BasePresenter
from mage_ai.orchestration.backfills.service import preview_run_dates


class BackfillPresenter(BasePresenter):
    default_attributes = [
        'block_uuid',
        'completed_at',
        'created_at',
        'end_datetime',
        'failed_at',
        'id',
        'interval_type',
        'interval_units',
        'metrics',
        'name',
        'pipeline_schedule_id',
        'pipeline_uuid',
        'start_datetime',
        'started_at',
        'status',
        'updated_at',
        'variables',
    ]

    def present(self, **kwargs):
        query = kwargs.get('query', {})

        data = self.model.to_dict()

        include_preview_runs = query.get('include_preview_runs', [False])
        include_run_count = query.get('include_run_count', [False])
        if include_preview_runs:
            include_preview_runs = include_preview_runs[0]
        if include_run_count:
            include_run_count = include_run_count[0]
        if (
            (include_preview_runs or include_run_count) and
            self.model.start_datetime is not None and
            self.model.end_datetime is not None and
            self.model.interval_type is not None and
            self.model.interval_units is not None
        ):
            pipeline_run_dates = preview_run_dates(self.model)
            data['total_run_count'] = len(pipeline_run_dates)
            if include_preview_runs:
                data['pipeline_run_dates'] = pipeline_run_dates

        return data



from unittest.mock import MagicMock, patch

from mage_ai.data_preparation.models.block import Block
from mage_ai.data_preparation.models.constants import PipelineType
from mage_ai.data_preparation.preferences import get_preferences
from mage_ai.data_preparation.sync.git_sync import GitSync
from mage_ai.data_preparation.variable_manager import VariableManager
from mage_ai.orchestration.db.models.schedules import BlockRun, PipelineRun
from mage_ai.orchestration.job_manager import JobType
from mage_ai.orchestration.pipeline_scheduler import PipelineScheduler
from mage_ai.tests.base_test import DBTestCase
from mage_ai.tests.factory import (
    create_pipeline_run_with_schedule,
    create_pipeline_with_blocks,
    create_pipeline_with_dynamic_blocks,
)


class PipelineSchedulerTests(DBTestCase):
    @classmethod
    def setUpClass(self):
        super().setUpClass()
        self.pipeline = create_pipeline_with_blocks(
            'test pipeline',
            self.repo_path,
        )
        self.dynamic_pipeline = create_pipeline_with_dynamic_blocks(
            'test dynamic pipeline',
            self.repo_path,
        )

    def test_start(self):
        pipeline_run = PipelineRun.create(pipeline_uuid='test_pipeline')
        scheduler = PipelineScheduler(pipeline_run=pipeline_run)
        with patch.object(scheduler, 'schedule') as mock_schedule:
            scheduler.start(should_schedule=False)
            self.assertEqual(mock_schedule.call_count, 0)
            self.assertEqual(pipeline_run.status, PipelineRun.PipelineRunStatus.RUNNING)

        pipeline_run2 = PipelineRun.create(pipeline_uuid='test_pipeline')
        scheduler2 = PipelineScheduler(pipeline_run=pipeline_run2)
        with patch.object(scheduler2, 'schedule') as mock_schedule2:
            scheduler2.start(should_schedule=True)
            mock_schedule2.assert_called_once()
            self.assertEqual(pipeline_run2.status, PipelineRun.PipelineRunStatus.RUNNING)

    def test_stop(self):
        pipeline_run = PipelineRun.create(pipeline_uuid='test_pipeline')
        block_runs = pipeline_run.block_runs
        for b in block_runs:
            b.update(status=BlockRun.BlockRunStatus.RUNNING)
        scheduler = PipelineScheduler(pipeline_run=pipeline_run)
        scheduler.stop()
        pipeline_run.refresh()
        block_runs = pipeline_run.block_runs
        for b in block_runs:
            self.assertEqual(b.status, BlockRun.BlockRunStatus.CANCELLED)

    @patch('mage_ai.orchestration.pipeline_scheduler.run_block')
    @patch('mage_ai.orchestration.pipeline_scheduler.job_manager')
    def test_schedule(self, mock_job_manager, mock_run_pipeline):
        pipeline_run = create_pipeline_run_with_schedule(pipeline_uuid='test_pipeline')
        scheduler = PipelineScheduler(pipeline_run=pipeline_run)
        mock_job_manager.add_job = MagicMock()
        scheduler.schedule()
        # TODO (tommy dang): change to 2 when we resume running heartbeat in pipeline scheduler
        self.assertEqual(mock_job_manager.add_job.call_count, 1)
        for b in pipeline_run.block_runs:
            if b.block_uuid == 'block1':
                self.assertEqual(b.status, BlockRun.BlockRunStatus.QUEUED)
            else:
                self.assertEqual(b.status, BlockRun.BlockRunStatus.INITIAL)

    def test_schedule_all_blocks_completed(self):
        pipeline_run = PipelineRun.create(pipeline_uuid='test_pipeline')
        pipeline_run.update(status=PipelineRun.PipelineRunStatus.RUNNING)
        for b in pipeline_run.block_runs:
            b.update(status=BlockRun.BlockRunStatus.COMPLETED)
        scheduler = PipelineScheduler(pipeline_run=pipeline_run)
        with patch.object(
            scheduler.notification_sender,
            'send_pipeline_run_success_message'
        ) as mock_send_message:
            scheduler.schedule()
            self.assertEqual(pipeline_run.status, PipelineRun.PipelineRunStatus.COMPLETED)
            self.assertEqual(mock_send_message.call_count, 1)

    def test_schedule_all_blocks_completed_with_failures(self):
        pipeline_run = create_pipeline_run_with_schedule(
            pipeline_uuid='test_pipeline',
            pipeline_schedule_settings=dict(allow_blocks_to_fail=True),
        )
        pipeline_run.update(status=PipelineRun.PipelineRunStatus.RUNNING)
        ct = 0
        for b in pipeline_run.block_runs:
            if ct == 0:
                b.update(status=BlockRun.BlockRunStatus.FAILED)
            else:
                b.update(status=BlockRun.BlockRunStatus.UPSTREAM_FAILED)
        scheduler = PipelineScheduler(pipeline_run=pipeline_run)
        with patch.object(
            scheduler.notification_sender,
            'send_pipeline_run_failure_message'
        ) as mock_send_message:
            scheduler.schedule()
            self.assertEqual(pipeline_run.status, PipelineRun.PipelineRunStatus.FAILED)
            self.assertEqual(mock_send_message.call_count, 1)

    @patch('mage_ai.orchestration.pipeline_scheduler.run_pipeline')
    @patch('mage_ai.orchestration.pipeline_scheduler.job_manager')
    def test_schedule_streaming(self, mock_job_manager, mock_run_pipeline):
        pipeline = create_pipeline_with_blocks(
            'test pipeline 2',
            self.repo_path,
        )
        pipeline.type = PipelineType.STREAMING
        pipeline.save()
        pipeline_run = create_pipeline_run_with_schedule(pipeline_uuid='test_pipeline_2')
        pipeline_run.update(status=PipelineRun.PipelineRunStatus.RUNNING)
        scheduler = PipelineScheduler(pipeline_run=pipeline_run)
        mock_has_pipeline_run_job = MagicMock()
        mock_job_manager.has_pipeline_run_job = mock_has_pipeline_run_job
        mock_has_pipeline_run_job.return_value = False
        mock_job_manager.add_job = MagicMock()
        scheduler.schedule()
        mock_job_manager.add_job.assert_called_once_with(
            JobType.PIPELINE_RUN,
            pipeline_run.id,
            mock_run_pipeline,
            pipeline_run.id,
            dict(
                env='prod',
                execution_date=None,
                execution_partition=f'{pipeline_run.pipeline_schedule_id}',
                event={},
            ),
            dict(
                pipeline_run_id=pipeline_run.id,
                pipeline_schedule_id=pipeline_run.pipeline_schedule_id,
                pipeline_uuid='test_pipeline_2',
            ),
        )

    def test_on_block_complete(self):
        pipeline_run = create_pipeline_run_with_schedule(pipeline_uuid='test_pipeline')
        pipeline_run.update(status=PipelineRun.PipelineRunStatus.RUNNING)
        scheduler = PipelineScheduler(pipeline_run=pipeline_run)
        with patch.object(scheduler, 'schedule') as mock_schedule:
            scheduler.on_block_complete('block1')
            mock_schedule.assert_called_once()
            block_run = BlockRun.get(pipeline_run_id=pipeline_run.id, block_uuid='block1')
            self.assertEqual(block_run.status, BlockRun.BlockRunStatus.COMPLETED)

    def test_on_block_failure(self):
        pipeline_run = create_pipeline_run_with_schedule(pipeline_uuid='test_pipeline')
        scheduler = PipelineScheduler(pipeline_run=pipeline_run)
        with patch.object(
            scheduler.notification_sender,
            'send_pipeline_run_failure_message'
        ) as mock_send_message:
            scheduler.on_block_failure('block1')
            mock_send_message.assert_called_once()
            block_run = BlockRun.get(pipeline_run_id=pipeline_run.id, block_uuid='block1')
            self.assertEqual(block_run.status, BlockRun.BlockRunStatus.FAILED)
            self.assertEqual(pipeline_run.status, PipelineRun.PipelineRunStatus.FAILED)

    def test_on_block_failure_allow_blocks_to_fail(self):
        pipeline_run = create_pipeline_run_with_schedule(
            pipeline_uuid='test_pipeline',
            pipeline_schedule_settings=dict(allow_blocks_to_fail=True),
        )
        pipeline_run.update(status=PipelineRun.PipelineRunStatus.RUNNING)
        scheduler = PipelineScheduler(pipeline_run=pipeline_run)
        with patch.object(
            scheduler.notification_sender,
            'send_pipeline_run_failure_message'
        ) as mock_send_message:
            scheduler.on_block_failure('block1')
            mock_send_message.assert_not_called()
            block_run = BlockRun.get(pipeline_run_id=pipeline_run.id, block_uuid='block1')
            self.assertEqual(block_run.status, BlockRun.BlockRunStatus.FAILED)
            self.assertEqual(pipeline_run.status, PipelineRun.PipelineRunStatus.RUNNING)

    # dynamic block tests
    @patch('mage_ai.orchestration.pipeline_scheduler.run_block')
    @patch('mage_ai.orchestration.pipeline_scheduler.job_manager')
    def test_schedule_for_dynamic_blocks(self, mock_job_manager, mock_run_pipeline):
        pipeline_run = create_pipeline_run_with_schedule(
            pipeline_uuid='test_dynamic_pipeline',
            pipeline_schedule_settings=dict(allow_blocks_to_fail=True),
        )
        scheduler = PipelineScheduler(pipeline_run=pipeline_run)
        scheduler.schedule()
        for b in pipeline_run.block_runs:
            if b.block_uuid == 'block1':
                self.assertEqual(b.status, BlockRun.BlockRunStatus.QUEUED)
                b.update(status=BlockRun.BlockRunStatus.COMPLETED)
            else:
                self.assertEqual(b.status, BlockRun.BlockRunStatus.INITIAL)
        with patch.object(scheduler, 'schedule') as mock_schedule:
            with patch.object(Block, 'output_variables') as mock_block_variables:
                with patch.object(VariableManager, 'get_variable') as mock_variable:
                    mock_block_variables.return_value = ['values', 'metadata']
                    # only mock the metadata
                    mock_variable.return_value = [
                        {
                            'block_uuid': 'for_user_1'
                        },
                        {
                            'block_uuid': 'for_user_2'
                        }
                    ]
                    scheduler.on_block_complete_without_schedule('block1')
                    mock_schedule.assert_not_called()
                    block_run1 = BlockRun.get(
                        pipeline_run_id=pipeline_run.id,
                        block_uuid='block2:for_user_1',
                    )
                    self.assertTrue(block_run1 is not None)
                    block_run2 = BlockRun.get(
                        pipeline_run_id=pipeline_run.id,
                        block_uuid='block2:for_user_2',
                    )
                    self.assertTrue(block_run2 is not None)

        scheduler.schedule()
        for b in pipeline_run.block_runs:
            if b.block_uuid.startswith('block2'):
                self.assertEqual(b.status, BlockRun.BlockRunStatus.QUEUED)

    @patch('mage_ai.orchestration.pipeline_scheduler.run_block')
    @patch('mage_ai.orchestration.pipeline_scheduler.job_manager')
    def test_schedule_for_dynamic_blocks_allow_blocks_to_fail(
        self,
        mock_job_manager,
        mock_run_pipeline,
    ):
        pipeline_run = create_pipeline_run_with_schedule(
            pipeline_uuid='test_dynamic_pipeline',
            pipeline_schedule_settings=dict(allow_blocks_to_fail=True),
        )
        scheduler = PipelineScheduler(pipeline_run=pipeline_run)
        scheduler.schedule()
        for b in pipeline_run.block_runs:
            if b.block_uuid == 'block1':
                self.assertEqual(b.status, BlockRun.BlockRunStatus.QUEUED)
                b.update(status=BlockRun.BlockRunStatus.COMPLETED)
            else:
                self.assertEqual(b.status, BlockRun.BlockRunStatus.INITIAL)
        with patch.object(scheduler, 'schedule') as mock_schedule:
            with patch.object(Block, 'output_variables') as mock_block_variables:
                with patch.object(VariableManager, 'get_variable') as mock_variable:
                    mock_block_variables.return_value = ['values', 'metadata']
                    # only mock the metadata
                    mock_variable.return_value = [
                        {
                            'block_uuid': 'for_user_1'
                        },
                        {
                            'block_uuid': 'for_user_2'
                        }
                    ]
                    scheduler.on_block_complete_without_schedule('block1')
                    mock_schedule.assert_not_called()
                    block_run1 = BlockRun.get(
                        pipeline_run_id=pipeline_run.id,
                        block_uuid='block2:for_user_1',
                    )
                    self.assertTrue(block_run1 is not None)
                    block_run2 = BlockRun.get(
                        pipeline_run_id=pipeline_run.id,
                        block_uuid='block2:for_user_2',
                    )
                    self.assertTrue(block_run2 is not None)

        scheduler.schedule()
        for b in pipeline_run.block_runs:
            if b.block_uuid.startswith('block2'):
                self.assertEqual(b.status, BlockRun.BlockRunStatus.QUEUED)

    def test_on_block_complete_dynamic_blocks(self):
        pipeline_run = create_pipeline_run_with_schedule(
            pipeline_uuid='test_dynamic_pipeline',
            pipeline_schedule_settings=dict(allow_blocks_to_fail=True),
        )
        scheduler = PipelineScheduler(pipeline_run=pipeline_run)
        with patch.object(scheduler, 'schedule') as mock_schedule:
            with patch.object(Block, 'output_variables') as mock_block_variables:
                with patch.object(VariableManager, 'get_variable') as mock_variable:
                    mock_block_variables.return_value = ['values', 'metadata']
                    # only mock the metadata
                    mock_variable.return_value = [
                        {
                            'block_uuid': 'for_user_1'
                        },
                        {
                            'block_uuid': 'for_user_2'
                        }
                    ]
                    scheduler.on_block_complete_without_schedule('block1')
                    mock_schedule.assert_not_called()
                    block_run1 = BlockRun.get(
                        pipeline_run_id=pipeline_run.id,
                        block_uuid='block2:for_user_1',
                    )
                    self.assertTrue(block_run1 is not None)
                    block_run2 = BlockRun.get(
                        pipeline_run_id=pipeline_run.id,
                        block_uuid='block2:for_user_2',
                    )
                    self.assertTrue(block_run2 is not None)

    def test_sync_data_on_pipeline_run(self):
        pipeline_run = create_pipeline_run_with_schedule(
            pipeline_uuid='test_dynamic_pipeline',
            pipeline_schedule_settings=dict(allow_blocks_to_fail=True),
        )
        scheduler = PipelineScheduler(pipeline_run=pipeline_run)
        preferences = get_preferences(repo_path=self.repo_path)
        preferences.update_preferences(
            dict(sync_config=dict(
                remote_repo_link='test_git_repo',
                repo_path=self.repo_path,
                branch='main',
                sync_on_pipeline_run=True,
            )))
        with patch.object(GitSync, 'sync_data') as mock_sync:
            scheduler.start(should_schedule=False)
            mock_sync.assert_called_once()



from mage_ai.api.oauth_scope import OauthScope
from mage_ai.api.operations import constants
from mage_ai.api.policies.BasePolicy import BasePolicy
from mage_ai.api.presenters.PipelineSchedulePresenter import PipelineSchedulePresenter
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.orchestration.db.models.oauth import Permission


class PipelineSchedulePolicy(BasePolicy):
    @property
    def entity(self):
        if self.resource and self.resource.model:
            return Permission.Entity.PIPELINE, self.resource.model.pipeline_uuid

        return Permission.Entity.PROJECT, get_repo_path()


PipelineSchedulePolicy.allow_actions([
    constants.DETAIL,
    constants.LIST,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_viewer_role())

PipelineSchedulePolicy.allow_actions([
    constants.CREATE,
    constants.DELETE,
    constants.UPDATE,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_editor_role())

PipelineSchedulePolicy.allow_read(PipelineSchedulePresenter.default_attributes + [], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.CREATE,
    constants.DELETE,
], condition=lambda policy: policy.has_at_least_editor_role())

PipelineSchedulePolicy.allow_read(PipelineSchedulePresenter.default_attributes + [
    'event_matchers',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.UPDATE,
], condition=lambda policy: policy.has_at_least_editor_role())

PipelineSchedulePolicy.allow_read(PipelineSchedulePresenter.default_attributes + [
    'event_matchers',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.DETAIL,
], condition=lambda policy: policy.has_at_least_viewer_role())

PipelineSchedulePolicy.allow_read(PipelineSchedulePresenter.default_attributes + [
    'event_matchers',
    'last_pipeline_run_status',
    'pipeline_runs_count',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.LIST,
], condition=lambda policy: policy.has_at_least_viewer_role())

PipelineSchedulePolicy.allow_write([
    'name',
    'repo_path',
    'schedule_interval',
    'schedule_type',
    'start_time',
    'status',
    'variables',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.CREATE,
], condition=lambda policy: policy.has_at_least_editor_role())

PipelineSchedulePolicy.allow_write([
    'event_matchers',
    'id',
    'name',
    'repo_path',
    'schedule_interval',
    'schedule_type',
    'settings',
    'sla',
    'start_time',
    'status',
    'variables',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.UPDATE,
], condition=lambda policy: policy.has_at_least_editor_role())

PipelineSchedulePolicy.allow_query([
    'order_by',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_viewer_role())



import asyncio
import json
import os
from unittest.mock import patch

import yaml

from mage_ai.data_preparation.models.block import Block
from mage_ai.data_preparation.models.constants import PipelineType
from mage_ai.data_preparation.models.pipeline import InvalidPipelineError, Pipeline
from mage_ai.data_preparation.models.widget import Widget
from mage_ai.tests.base_test import DBTestCase
from mage_ai.tests.factory import create_pipeline_run_with_schedule


class PipelineTest(DBTestCase):
    def test_create(self):
        pipeline = Pipeline.create(
            'test pipeline',
            repo_path=self.repo_path,
        )
        self.assertEqual(pipeline.uuid, 'test_pipeline')
        self.assertEqual(pipeline.name, 'test pipeline')
        self.assertEqual(pipeline.blocks_by_uuid, dict())
        self.assertTrue(os.path.exists(f'{self.repo_path}/pipelines/test_pipeline/__init__.py'))
        self.assertTrue(os.path.exists(f'{self.repo_path}/pipelines/test_pipeline/metadata.yaml'))

    def test_add_block(self):
        self.__create_pipeline_with_blocks('test pipeline 2')
        pipeline = Pipeline('test_pipeline_2', self.repo_path)

        self.assertEqual(pipeline.to_dict(), dict(
            data_integration=None,
            description=None,
            executor_config=dict(),
            executor_count=1,
            executor_type=None,
            name='test pipeline 2',
            uuid='test_pipeline_2',
            spark_config=dict(),
            type='python',
            blocks=[
                dict(
                    language='python',
                    color=None,
                    configuration={},
                    executor_config=None,
                    executor_type='local_python',
                    has_callback=False,
                    name='block1',
                    uuid='block1',
                    type='data_loader',
                    retry_config=None,
                    status='not_executed',
                    upstream_blocks=[],
                    downstream_blocks=['block2', 'block3'],
                    all_upstream_blocks_executed=True,
                ),
                dict(
                    language='python',
                    color=None,
                    configuration={},
                    executor_config=None,
                    executor_type='local_python',
                    has_callback=False,
                    name='block2',
                    uuid='block2',
                    type='transformer',
                    retry_config=None,
                    status='not_executed',
                    upstream_blocks=['block1'],
                    downstream_blocks=['block4'],
                    all_upstream_blocks_executed=False,
                ),
                dict(
                    language='python',
                    color=None,
                    configuration={},
                    executor_config=None,
                    executor_type='local_python',
                    has_callback=False,
                    name='block3',
                    uuid='block3',
                    type='transformer',
                    retry_config=None,
                    status='not_executed',
                    upstream_blocks=['block1'],
                    downstream_blocks=['block4'],
                    all_upstream_blocks_executed=False,
                ),
                dict(
                    language='python',
                    color=None,
                    configuration={},
                    executor_config=None,
                    executor_type='local_python',
                    has_callback=False,
                    name='block4',
                    uuid='block4',
                    type='data_exporter',
                    retry_config=None,
                    status='not_executed',
                    upstream_blocks=['block2', 'block3'],
                    downstream_blocks=['widget1'],
                    all_upstream_blocks_executed=False,
                ),
            ],
            callbacks=[],
            conditionals=[],
            updated_at=None,
            widgets=[
                dict(
                    language='python',
                    color=None,
                    configuration={},
                    executor_config=None,
                    executor_type='local_python',
                    has_callback=False,
                    name='widget1',
                    uuid='widget1',
                    type='chart',
                    retry_config=None,
                    status='not_executed',
                    upstream_blocks=['block4'],
                    downstream_blocks=[],
                    all_upstream_blocks_executed=False,
                ),
            ],
        ))

    def test_update_name_existing_pipeline(self):
        pipeline1 = Pipeline.create(
            'test_pipeline_a',
            repo_path=self.repo_path,
        )
        pipeline2 = Pipeline.create(
            'test_pipeline_b',
            repo_path=self.repo_path,
        )
        with self.assertRaisesRegex(Exception, 'Pipeline test_pipeline_a already exists.'):
            asyncio.run(pipeline2.update(dict(name='test_pipeline_a', uuid='test_pipeline_b')))
        self.assertEqual(pipeline1.name, 'test_pipeline_a')
        self.assertEqual(pipeline2.name, 'test_pipeline_b')

    def test_update_name(self):
        pipeline = Pipeline.create(
            'test_pipeline_c',
            repo_path=self.repo_path,
        )
        pipeline_run = create_pipeline_run_with_schedule(pipeline_uuid='test_pipeline_c')
        pipeline_schedule = pipeline_run.pipeline_schedule
        self.assertEqual(pipeline.name, 'test_pipeline_c')
        self.assertEqual(pipeline.uuid, 'test_pipeline_c')
        self.assertEqual(pipeline_run.pipeline_uuid, 'test_pipeline_c')
        self.assertEqual(pipeline_schedule.pipeline_uuid, 'test_pipeline_c')
        asyncio.run(pipeline.update(dict(name='test_pipeline_c2', uuid='test_pipeline_c')))
        self.assertEqual(pipeline.name, 'test_pipeline_c2')
        self.assertEqual(pipeline.uuid, 'test_pipeline_c2')
        self.assertEqual(pipeline_run.pipeline_uuid, 'test_pipeline_c2')
        self.assertEqual(pipeline_schedule.pipeline_uuid, 'test_pipeline_c2')

    def test_delete_block(self):
        pipeline = self.__create_pipeline_with_blocks('test pipeline 3')
        block = pipeline.blocks_by_uuid['block4']
        widget = pipeline.widgets_by_uuid['widget1']
        pipeline.delete_block(widget, widget=True)
        pipeline.delete_block(block)
        pipeline = Pipeline('test_pipeline_3', self.repo_path)
        self.assertEqual(pipeline.to_dict(), dict(
            data_integration=None,
            description=None,
            executor_config=dict(),
            executor_count=1,
            executor_type=None,
            name='test pipeline 3',
            uuid='test_pipeline_3',
            spark_config=dict(),
            type='python',
            blocks=[
                dict(
                    language='python',
                    color=None,
                    configuration={},
                    executor_config=None,
                    executor_type='local_python',
                    has_callback=False,
                    name='block1',
                    uuid='block1',
                    type='data_loader',
                    retry_config=None,
                    status='not_executed',
                    upstream_blocks=[],
                    downstream_blocks=['block2', 'block3'],
                    all_upstream_blocks_executed=True,
                ),
                dict(
                    language='python',
                    color=None,
                    configuration={},
                    executor_config=None,
                    executor_type='local_python',
                    has_callback=False,
                    name='block2',
                    uuid='block2',
                    type='transformer',
                    retry_config=None,
                    status='not_executed',
                    upstream_blocks=['block1'],
                    downstream_blocks=[],
                    all_upstream_blocks_executed=False,
                ),
                dict(
                    language='python',
                    color=None,
                    configuration={},
                    executor_config=None,
                    executor_type='local_python',
                    has_callback=False,
                    name='block3',
                    uuid='block3',
                    type='transformer',
                    retry_config=None,
                    status='not_executed',
                    upstream_blocks=['block1'],
                    downstream_blocks=[],
                    all_upstream_blocks_executed=False,
                )
            ],
            callbacks=[],
            conditionals=[],
            updated_at=None,
            widgets=[],
        ))

    def test_execute(self):
        pipeline = Pipeline.create(
            'test pipeline 4',
            repo_path=self.repo_path,
        )
        block1 = self.__create_dummy_data_loader_block('block1', pipeline)
        block2 = self.__create_dummy_transformer_block('block2', pipeline)
        block3 = self.__create_dummy_transformer_block('block3', pipeline)
        block4 = self.__create_dummy_data_exporter_block('block4', pipeline)
        pipeline.add_block(block1)
        pipeline.add_block(block2, upstream_block_uuids=['block1'])
        pipeline.add_block(block3, upstream_block_uuids=['block1'])
        pipeline.add_block(block4, upstream_block_uuids=['block2', 'block3'])
        pipeline.execute_sync()
        self.assertEqual(pipeline.to_dict(), dict(
            data_integration=None,
            description=None,
            executor_config=dict(),
            executor_count=1,
            executor_type=None,
            name='test pipeline 4',
            uuid='test_pipeline_4',
            spark_config=dict(),
            type='python',
            blocks=[
                dict(
                    language='python',
                    color=None,
                    configuration={},
                    executor_config=None,
                    executor_type='local_python',
                    has_callback=False,
                    name='block1',
                    uuid='block1',
                    type='data_loader',
                    retry_config=None,
                    status='executed',
                    upstream_blocks=[],
                    downstream_blocks=['block2', 'block3'],
                    all_upstream_blocks_executed=True,
                ),
                dict(
                    language='python',
                    color=None,
                    configuration={},
                    executor_config=None,
                    executor_type='local_python',
                    has_callback=False,
                    name='block2',
                    uuid='block2',
                    type='transformer',
                    retry_config=None,
                    status='executed',
                    upstream_blocks=['block1'],
                    downstream_blocks=['block4'],
                    all_upstream_blocks_executed=True,
                ),
                dict(
                    language='python',
                    color=None,
                    configuration={},
                    executor_config=None,
                    executor_type='local_python',
                    has_callback=False,
                    name='block3',
                    uuid='block3',
                    type='transformer',
                    retry_config=None,
                    status='executed',
                    upstream_blocks=['block1'],
                    downstream_blocks=['block4'],
                    all_upstream_blocks_executed=True,
                ),
                dict(
                    language='python',
                    color=None,
                    configuration={},
                    executor_config=None,
                    executor_type='local_python',
                    has_callback=False,
                    name='block4',
                    uuid='block4',
                    type='data_exporter',
                    retry_config=None,
                    status='executed',
                    upstream_blocks=['block2', 'block3'],
                    downstream_blocks=[],
                    all_upstream_blocks_executed=True,
                )
            ],
            callbacks=[],
            conditionals=[],
            updated_at=None,
            widgets=[],
        ))

    def test_execute_multiple_paths(self):
        pipeline = Pipeline.create(
            'test pipeline 5',
            repo_path=self.repo_path,
        )
        block1 = self.__create_dummy_data_loader_block('block1', pipeline)
        block2 = self.__create_dummy_transformer_block('block2', pipeline)
        block3 = self.__create_dummy_transformer_block('block3', pipeline)
        block4 = self.__create_dummy_data_loader_block('block4', pipeline)
        block5 = self.__create_dummy_transformer_block('block5', pipeline)
        block6 = self.__create_dummy_transformer_block('block6', pipeline)
        block7 = self.__create_dummy_data_exporter_block('block7', pipeline)
        pipeline.add_block(block1)
        pipeline.add_block(block2, upstream_block_uuids=['block1'])
        pipeline.add_block(block3, upstream_block_uuids=['block1'])
        pipeline.add_block(block4)
        pipeline.add_block(block5, upstream_block_uuids=['block4'])
        pipeline.add_block(block6, upstream_block_uuids=['block5'])
        pipeline.add_block(block7, upstream_block_uuids=['block2', 'block3', 'block6'])
        pipeline.execute_sync()
        self.assertEqual(pipeline.to_dict(), dict(
            data_integration=None,
            description=None,
            executor_config=dict(),
            executor_count=1,
            executor_type=None,
            name='test pipeline 5',
            uuid='test_pipeline_5',
            spark_config=dict(),
            type='python',
            blocks=[
                dict(
                    language='python',
                    color=None,
                    configuration={},
                    executor_config=None,
                    executor_type='local_python',
                    has_callback=False,
                    name='block1',
                    uuid='block1',
                    type='data_loader',
                    retry_config=None,
                    status='executed',
                    upstream_blocks=[],
                    downstream_blocks=['block2', 'block3'],
                    all_upstream_blocks_executed=True,
                ),
                dict(
                    language='python',
                    color=None,
                    configuration={},
                    executor_config=None,
                    executor_type='local_python',
                    has_callback=False,
                    name='block2',
                    uuid='block2',
                    type='transformer',
                    retry_config=None,
                    status='executed',
                    upstream_blocks=['block1'],
                    downstream_blocks=['block7'],
                    all_upstream_blocks_executed=True,
                ),
                dict(
                    language='python',
                    color=None,
                    configuration={},
                    executor_config=None,
                    executor_type='local_python',
                    has_callback=False,
                    name='block3',
                    uuid='block3',
                    type='transformer',
                    retry_config=None,
                    status='executed',
                    upstream_blocks=['block1'],
                    downstream_blocks=['block7'],
                    all_upstream_blocks_executed=True,
                ),
                dict(
                    language='python',
                    color=None,
                    configuration={},
                    executor_config=None,
                    executor_type='local_python',
                    has_callback=False,
                    name='block4',
                    uuid='block4',
                    type='data_loader',
                    retry_config=None,
                    status='executed',
                    upstream_blocks=[],
                    downstream_blocks=['block5'],
                    all_upstream_blocks_executed=True,
                ),
                dict(
                    language='python',
                    color=None,
                    configuration={},
                    executor_config=None,
                    executor_type='local_python',
                    has_callback=False,
                    name='block5',
                    uuid='block5',
                    type='transformer',
                    retry_config=None,
                    status='executed',
                    upstream_blocks=['block4'],
                    downstream_blocks=['block6'],
                    all_upstream_blocks_executed=True,
                ),
                dict(
                    language='python',
                    color=None,
                    configuration={},
                    executor_config=None,
                    executor_type='local_python',
                    has_callback=False,
                    name='block6',
                    uuid='block6',
                    type='transformer',
                    retry_config=None,
                    status='executed',
                    upstream_blocks=['block5'],
                    downstream_blocks=['block7'],
                    all_upstream_blocks_executed=True,
                ),
                dict(
                    language='python',
                    color=None,
                    configuration={},
                    executor_config=None,
                    executor_type='local_python',
                    has_callback=False,
                    name='block7',
                    uuid='block7',
                    type='data_exporter',
                    retry_config=None,
                    status='executed',
                    upstream_blocks=['block2', 'block3', 'block6'],
                    downstream_blocks=[],
                    all_upstream_blocks_executed=True,
                )
            ],
            callbacks=[],
            conditionals=[],
            updated_at=None,
            widgets=[],
        ))

    def test_delete(self):
        pipeline = Pipeline.create(
            'test pipeline 6',
            repo_path=self.repo_path,
        )
        block1 = self.__create_dummy_data_loader_block('block1', pipeline)
        block2 = self.__create_dummy_transformer_block('block2', pipeline)
        block3 = self.__create_dummy_data_exporter_block('block3', pipeline)
        block4 = self.__create_dummy_scratchpad('block4', pipeline)
        block5 = self.__create_dummy_scratchpad('block5', pipeline)
        pipeline.add_block(block1)
        pipeline.add_block(block2, upstream_block_uuids=['block1'])
        pipeline.add_block(block3, upstream_block_uuids=['block2'])
        pipeline.add_block(block4)
        pipeline.add_block(block5)
        pipeline.delete()
        self.assertFalse(os.access(pipeline.dir_path, os.F_OK))
        self.assertTrue(os.access(block1.file_path, os.F_OK))
        self.assertTrue(os.access(block2.file_path, os.F_OK))
        self.assertTrue(os.access(block3.file_path, os.F_OK))
        self.assertFalse(os.access(block4.file_path, os.F_OK))
        self.assertFalse(os.access(block5.file_path, os.F_OK))

    def test_duplicate_standard_pipeline(self):
        pipeline = self.__create_pipeline_with_blocks('test_pipeline_7a')
        duplicate_pipeline = Pipeline.duplicate(pipeline, 'duplicate_pipeline')
        for block_uuid in pipeline.blocks_by_uuid:
            original = pipeline.blocks_by_uuid[block_uuid]
            duplicate = duplicate_pipeline.blocks_by_uuid[block_uuid]
            self.assertEqual(original.name, duplicate.name)
            self.assertEqual(original.uuid, duplicate.uuid)
            self.assertEqual(original.type, duplicate.type)
            self.assertEqual(original.upstream_block_uuids, duplicate.upstream_block_uuids)
            self.assertEqual(original.downstream_block_uuids, duplicate.downstream_block_uuids)
        for widget_uuid in pipeline.widgets_by_uuid:
            original = pipeline.widgets_by_uuid[widget_uuid]
            duplicate = duplicate_pipeline.widgets_by_uuid[widget_uuid]
            self.assertEqual(original.name, duplicate.name)
            self.assertEqual(original.uuid, duplicate.uuid)
            self.assertEqual(original.type, duplicate.type)
            self.assertEqual(original.chart_type, duplicate.chart_type)
            self.assertEqual(original.upstream_block_uuids, duplicate.upstream_block_uuids)

    def test_duplicate_integration_pipeline(self):
        pipeline = self.__create_pipeline_with_integration('test_pipeline_7b')
        duplicate_pipeline = Pipeline.duplicate(pipeline, 'duplicate_pipeline_2')
        for block_uuid in pipeline.blocks_by_uuid:
            original = pipeline.blocks_by_uuid[block_uuid]
            duplicate = duplicate_pipeline.blocks_by_uuid[block_uuid]
            self.assertEqual(original.name, duplicate.name)
            self.assertEqual(original.uuid, duplicate.uuid)
            self.assertEqual(original.type, duplicate.type)
            self.assertEqual(original.upstream_block_uuids, duplicate.upstream_block_uuids)
            self.assertEqual(original.downstream_block_uuids, duplicate.downstream_block_uuids)
            self.assertEqual(pipeline.data_integration, duplicate_pipeline.data_integration)

    def test_cycle_detection(self):
        pipeline = self.__create_pipeline_with_blocks('test pipeline 8')
        pipeline.validate()

        block_new = Block.create('block_new', 'transformer', self.repo_path)
        block_new.downstream_blocks = pipeline.get_blocks(['block1', 'block2'])
        with self.assertRaises(InvalidPipelineError):
            pipeline.add_block(block_new, upstream_block_uuids=['block4'])
        block_new.downstream_blocks = pipeline.get_blocks(['block2'])
        with self.assertRaises(InvalidPipelineError):
            pipeline.add_block(block_new, upstream_block_uuids=['block4'])

        block4 = pipeline.get_block('block4')
        block4.downstream_blocks = pipeline.get_blocks(['block1', 'block2'])
        with self.assertRaises(InvalidPipelineError):
            pipeline.update_block(block4)

    def test_save_and_get_data_integration_catalog(self):
        pipeline = self.__create_pipeline_with_integration('test_pipeline_9')
        pipeline.save()
        catalog_config_path = os.path.join(
            self.repo_path,
            'pipelines/test_pipeline_9/data_integration_catalog.json',
        )
        self.assertEqual(pipeline.catalog_config_path, catalog_config_path)
        self.assertTrue(os.path.exists(catalog_config_path))
        expected_catalog_config = {
            'catalog': {
                'streams': [
                    {
                        'tap_stream_id': 'demo_users',
                        'stream': 'demo_users',
                    },
                ],
            }
        }
        with open(catalog_config_path) as f:
            catalog_json = json.load(f)
            self.assertEqual(catalog_json, expected_catalog_config)
        self.assertTrue(os.path.exists(pipeline.config_path))
        with open(pipeline.config_path) as f:
            config_json = yaml.full_load(f)
            self.assertEqual(
                config_json,
                {
                    "data_integration": None,
                    "description": None,
                    "executor_config": {},
                    "executor_count": 1,
                    "executor_type": None,
                    "extensions": {},
                    "name": "test_pipeline_9",
                    "spark_config": {},
                    "type": "integration",
                    "updated_at": None,
                    "uuid": "test_pipeline_9",
                    "blocks": [
                        {
                            "all_upstream_blocks_executed": True,
                            "color": None,
                            "configuration": {},
                            "downstream_blocks": ["destination_block"],
                            "executor_config": None,
                            "executor_type": "local_python",
                            "has_callback": False,
                            "name": "source_block",
                            "language": "python",
                            "retry_config": None,
                            "status": "not_executed",
                            "type": "data_loader",
                            "upstream_blocks": [],
                            "uuid": "source_block",
                        },
                        {
                            "all_upstream_blocks_executed": False,
                            "color": None,
                            "configuration": {},
                            "downstream_blocks": [],
                            "executor_config": None,
                            "executor_type": "local_python",
                            "has_callback": False,
                            "name": "destination_block",
                            "language": "python",
                            "retry_config": None,
                            "status": "not_executed",
                            "type": "transformer",
                            "upstream_blocks": ["source_block"],
                            "uuid": "destination_block",
                        },
                    ],
                    "callbacks": [],
                    "conditionals": [],
                    "widgets": [],
                },
            )
        pipeline_load = Pipeline.get('test_pipeline_9')
        self.assertEqual(pipeline_load.to_dict()['data_integration'], expected_catalog_config)

    def test_save_and_get_integration_pipeline_async(self):
        pipeline = self.__create_pipeline_with_integration('test_pipeline_10')
        asyncio.run(pipeline.save_async())

        pipeline_load = asyncio.run(Pipeline.get_async('test_pipeline_10'))
        self.assertEqual(
            pipeline_load.to_dict()['data_integration'],
            {
                'catalog': {
                    'streams': [
                        {
                            'tap_stream_id': 'demo_users',
                            'stream': 'demo_users',
                        },
                    ],
                }
            },
        )
        self.assertEqual(
            pipeline_load.to_dict(),
            pipeline.to_dict(),
        )

    def test_save_with_empty_content(self):
        pipeline = self.__create_pipeline_with_blocks('test pipeline 11')
        with patch.object(pipeline, 'to_dict', return_value=dict()):
            with self.assertRaises(Exception) as err:
                pipeline.save()
            self.assertTrue('Writing empty pipeline metadata is prevented.' in str(err.exception))

    def __create_pipeline_with_blocks(self, name):
        pipeline = Pipeline.create(
            name,
            repo_path=self.repo_path,
        )
        block1 = Block.create('block1', 'data_loader', self.repo_path, language='python')
        block2 = Block.create('block2', 'transformer', self.repo_path, language='python')
        block3 = Block.create('block3', 'transformer', self.repo_path, language='python')
        block4 = Block.create('block4', 'data_exporter', self.repo_path, language='python')
        widget1 = Widget.create('widget1', 'chart', self.repo_path, language='python')
        pipeline.add_block(block1)
        pipeline.add_block(block2, upstream_block_uuids=['block1'])
        pipeline.add_block(block3, upstream_block_uuids=['block1'])
        pipeline.add_block(block4, upstream_block_uuids=['block2', 'block3'])
        pipeline.add_block(widget1, upstream_block_uuids=['block4'], widget=True)
        return pipeline

    def __create_pipeline_with_integration(self, name):
        pipeline = Pipeline.create(
            name,
            pipeline_type=PipelineType.INTEGRATION,
            repo_path=self.repo_path,
        )
        source_block = Block.create(
            'source_block',
            'data_loader',
            self.repo_path,
            language='python',
        )
        destination_block = Block.create(
            'destination_block',
            'transformer',
            self.repo_path,
            language='python',
        )
        pipeline.add_block(source_block)
        pipeline.add_block(destination_block, upstream_block_uuids=['source_block'])
        pipeline.data_integration = {
            'catalog': {
                'streams': [
                    {
                        'tap_stream_id': 'demo_users',
                        'stream': 'demo_users',
                    },
                ],
            },
        }
        return pipeline

    def __create_dummy_data_loader_block(self, name, pipeline):
        block = Block.create(
            name,
            'data_loader',
            self.repo_path,
            pipeline=pipeline,
            language='python',
        )
        with open(block.file_path, 'w') as file:
            file.write('''import pandas as pd
@data_loader
def load_data():
    data = {'col1': [1, 1, 3], 'col2': [2, 2, 4]}
    df = pd.DataFrame(data)
    return [df]
            ''')
        return block

    def __create_dummy_transformer_block(self, name, pipeline):
        block = Block.create(
            name,
            'transformer',
            self.repo_path,
            pipeline=pipeline,
            language='python',
        )
        with open(block.file_path, 'w') as file:
            file.write('''import pandas as pd
@transformer
def transform(df):
    return df
            ''')
        return block

    def __create_dummy_data_exporter_block(self, name, pipeline):
        block = Block.create(
            name,
            'data_exporter',
            self.repo_path,
            pipeline=pipeline,
            language='python',
        )
        with open(block.file_path, 'w') as file:
            file.write('''import pandas as pd
@data_exporter
def export_data(df, *args):
    return None
            ''')
        return block

    def __create_dummy_scratchpad(self, name, pipeline):
        block = Block.create(
            name,
            'scratchpad',
            self.repo_path,
            pipeline=pipeline,
            language='python',
        )
        with open(block.file_path, 'w') as file:
            file.write(
                '''import antigravity
            '''
            )
        return block



from mage_ai.api.oauth_scope import OauthScope
from mage_ai.api.operations import constants
from mage_ai.api.policies.BasePolicy import BasePolicy
from mage_ai.api.presenters.IntegrationDestinationPresenter import IntegrationDestinationPresenter


class IntegrationDestinationPolicy(BasePolicy):
    pass


IntegrationDestinationPolicy.allow_actions([
    constants.CREATE,
    constants.LIST,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_viewer_role())

IntegrationDestinationPolicy.allow_read(IntegrationDestinationPresenter.default_attributes + [
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.LIST,
], condition=lambda policy: policy.has_at_least_viewer_role())

IntegrationDestinationPolicy.allow_read([
    'error_message',
    'success',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.CREATE,
], condition=lambda policy: policy.has_at_least_viewer_role())

IntegrationDestinationPolicy.allow_write([
    'action_type',
    'config',
    'pipeline_uuid',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.CREATE,
], condition=lambda policy: policy.has_at_least_viewer_role())



from dataclasses import dataclass, field
from mage_ai.shared.config import BaseConfig
from typing import List, Dict


@dataclass
class OpsgenieConfig(BaseConfig):
    url: str = None
    api_key: str = None
    priority: str = "P3"
    responders: List[Dict] = field(default_factory=list)
    tags: List = field(default_factory=list)
    details: Dict = field(default_factory=dict)

    @property
    def is_valid(self) -> bool:
        return self.url is not None and self.url != 'None'\
            and self.api_key is not None and self.api_key != 'None'



from mage_ai.services.spark.config import SparkConfig
import os


def has_same_spark_config(spark_session, spark_config: SparkConfig) -> bool:
    if spark_session is None:
        return False

    if spark_config:
        if (spark_config.app_name and spark_config.app_name
                != spark_session.conf.get('spark.app.name')):
            print('Different spark.app.name found!')
            return False
        if (spark_config.spark_master and spark_config.spark_master
                != spark_session.conf.get('spark.master')):
            print('Different spark.master found!')
            return False
        if (spark_config.spark_home and spark_config.spark_home
                != spark_session.conf.get('spark.home')):
            print('Different spark.home found!')
            return False
        if spark_config.executor_env:
            for key, value in spark_config.executor_env.items():
                if spark_session.conf.get(f'spark.executorEnv.{key}') != value:
                    print(f'Different spark.executorEnv.{key} found!')
                    return False
        if (spark_config.spark_jars and spark_config.spark_jars
                != spark_session.conf.get('spark.jars')):
            print('Different spark.jars found!')
            return False
        if spark_config.others:
            for key, value in spark_config.others.items():
                if spark_session.conf.get(key) != value:
                    print(f'Different {key} found!')
                    return False
    return True


def get_spark_session(spark_config: SparkConfig):
    from pyspark.conf import SparkConf
    from pyspark.sql import SparkSession

    if spark_config:
        active_session = SparkSession.getActiveSession()
        print('Check the given spark_config against the active Spark session.')
        if has_same_spark_config(
            spark_session=active_session,
            spark_config=spark_config
        ):
            print('Reuse the active Spark session.')
            return active_session
        else:
            print('Create a new Spark session.')
            if active_session:
                active_session.stop()
            conf = SparkConf()
            if spark_config.app_name:
                conf.setAppName(spark_config.app_name)
            if spark_config.spark_master:
                conf.setMaster(spark_config.spark_master)
            if spark_config.spark_home:
                conf.setSparkHome(spark_config.spark_home)
            if spark_config.executor_env:
                list_kv_pairs = []
                for key, value in spark_config.executor_env.items():
                    list_kv_pairs.append((key, value))
                conf.setExecutorEnv(key=None, value=None, pairs=list_kv_pairs)
            if spark_config.spark_jars:
                conf.set('spark.jars', ','.join(spark_config.spark_jars))
            if spark_config.others:
                conf.setAll(spark_config.others)

            return SparkSession.builder.config(conf=conf).getOrCreate()
    else:
        return SparkSession.builder.master(
            os.getenv('SPARK_MASTER_HOST', 'local')).getOrCreate()



import os
from typing import Dict, List

from kubernetes import client, config

from mage_ai.cluster_manager.constants import (
    CLOUD_SQL_CONNECTION_NAME,
    CONNECTION_URL_SECRETS_NAME,
    DB_SECRETS_NAME,
    GCP_BACKEND_CONFIG_ANNOTATION,
    KUBE_SERVICE_GCP_BACKEND_CONFIG,
    KUBE_SERVICE_TYPE,
    NODE_PORT_SERVICE_TYPE,
    SERVICE_ACCOUNT_CREDENTIAL_FILE_PATH,
    SERVICE_ACCOUNT_SECRETS_NAME,
)
from mage_ai.orchestration.constants import (
    DATABASE_CONNECTION_URL_ENV_VAR,
    DB_NAME,
    DB_PASS,
    DB_USER,
)


class WorkloadManager:
    def __init__(self, namespace: str = 'default'):
        self.load_config()
        self.core_client = client.CoreV1Api()
        self.apps_client = client.AppsV1Api()

        self.namespace = namespace
        if not self.namespace:
            self.namespace = 'default'

    @classmethod
    def load_config(cls) -> bool:
        try:
            config.load_incluster_config()
            return True
        except Exception:
            pass

        try:
            config.load_kube_config()
        except Exception:
            pass

        return False

    def list_services(self):
        services = self.core_client.list_namespaced_service(self.namespace).items
        services_list = []
        for service in services:
            try:
                labels = service.metadata.labels
                if not labels.get('dev-instance'):
                    continue
                ip_address = service.status.load_balancer.ingress[0].ip
                conditions = service.status.conditions or list()
                services_list.append(dict(
                    ip=ip_address,
                    name=labels.get('app'),
                    status='RUNNING' if len(conditions) == 0 else conditions[0].status,
                    type='kubernetes',
                ))
            except Exception:
                pass

        return services_list

    def create_deployment(
        self,
        deployment_name,
        container_config: Dict = None,
        service_account_name: str = None,
        storage_class_name: str = None,
        volume_host_path: str = None,
    ):
        if container_config is None:
            container_config = dict()

        env_vars = self.__populate_env_vars(container_config)
        container_config['env'] = env_vars

        containers = [
            {
                'name': f'{deployment_name}-container',
                'image': 'mageai/mageai:latest',
                'command': ['mage', 'start', deployment_name],
                'ports': [
                    {
                        'containerPort': 6789,
                        'name': 'web'
                    }
                ],
                'volumeMounts': [
                    {
                        'name': 'mage-data',
                        'mountPath': '/home/src'
                    }
                ],
                **container_config,
            }
        ]

        volumes = []
        if os.getenv(SERVICE_ACCOUNT_SECRETS_NAME):
            credential_file_path = os.getenv(
                SERVICE_ACCOUNT_CREDENTIAL_FILE_PATH,
                'service_account.json',
            )
            containers.append(
                {
                    'name': 'cloud-sql-proxy',
                    'image': 'gcr.io/cloudsql-docker/gce-proxy:1.28.0',
                    'command': [
                        '/cloud_sql_proxy',
                        '-log_debug_stdout',
                        f'-instances={os.getenv(CLOUD_SQL_CONNECTION_NAME)}=tcp:5432',
                        f'-credential_file=/secrets/{credential_file_path}'
                    ],
                    'securityContext': {
                        'runAsNonRoot': True
                    },
                    'resources': {
                        'requests': {
                            'memory': '1Gi',
                            'cpu': '1'
                        }
                    },
                    'volumeMounts': [
                        {
                            'name': 'service-account-volume',
                            'mountPath': '/secrets/',
                            'readOnly': True
                        }
                    ]
                }
            )
            volumes.append(
                {
                    'name': 'service-account-volume',
                    'secret': {
                        'secretName': os.getenv(SERVICE_ACCOUNT_SECRETS_NAME)
                    }
                }
            )

        if volume_host_path:
            volumes.append(
                {
                    'name': 'mage-data',
                    'hostPath': {
                        'path': volume_host_path
                    },
                }
            )
        deployment_template_spec = dict()
        if service_account_name:
            deployment_template_spec['serviceAccountName'] = service_account_name

        deployment = {
            'apiVersion': 'apps/v1',
            'kind': 'Deployment',
            'metadata': {
                'name': deployment_name,
                'labels': {
                    'app': deployment_name
                }
            },
            'spec': {
                'selector': {
                    'matchLabels': {
                        'app': deployment_name
                    }
                },
                'replicas': 1,
                'minReadySeconds': 10,
                'template': {
                    'metadata': {
                        'labels': {
                            'app': deployment_name
                        }
                    },
                    'spec': {
                        'terminationGracePeriodSeconds': 10,
                        'containers': containers,
                        'volumes': volumes,
                        **deployment_template_spec
                    }
                },
            }
        }

        self.apps_client.create_namespaced_deployment(self.namespace, deployment)

        service_name = f'{deployment_name}-service'

        annotations = {}
        if os.getenv(KUBE_SERVICE_GCP_BACKEND_CONFIG):
            annotations[GCP_BACKEND_CONFIG_ANNOTATION] = \
                os.getenv(KUBE_SERVICE_GCP_BACKEND_CONFIG)

        service = {
            'apiVersion': 'v1',
            'kind': 'Service',
            'metadata': {
                'name': service_name,
                'labels': {
                    'app': deployment_name,
                    'dev-instance': '1',
                },
                'annotations': annotations
            },
            'spec': {
                'ports': [
                    {
                        'protocol': 'TCP',
                        'port': 6789,
                    }
                ],
                'selector': {
                    'app': deployment_name
                },
                'type': os.getenv(KUBE_SERVICE_TYPE, NODE_PORT_SERVICE_TYPE)
            }
        }

        return self.core_client.create_namespaced_service(self.namespace, service)

    def __populate_env_vars(self, container_config) -> List:
        env_vars = []

        connection_url_secrets_name = os.getenv(CONNECTION_URL_SECRETS_NAME)
        if connection_url_secrets_name:
            env_vars.append(
                {
                    'name': DATABASE_CONNECTION_URL_ENV_VAR,
                    'valueFrom': {
                        'secretKeyRef': {
                            'name': connection_url_secrets_name,
                            'key': 'connection_url'
                        }
                    }
                }
            )

        # For connecting to CloudSQL PostgreSQL database.
        db_secrets_name = os.getenv(DB_SECRETS_NAME)
        if db_secrets_name:
            env_vars.extend([
                {
                    'name': DB_USER,
                    'valueFrom': {
                        'secretKeyRef': {
                            'name': db_secrets_name,
                            'key': 'username'
                        }
                    }
                },
                {
                    'name': DB_PASS,
                    'valueFrom': {
                        'secretKeyRef': {
                            'name': db_secrets_name,
                            'key': 'password'
                        }
                    }
                },
                {
                    'name': DB_NAME,
                    'valueFrom': {
                        'secretKeyRef': {
                            'name': db_secrets_name,
                            'key': 'database'
                        }
                    }
                }
            ])

        if container_config and 'env' in container_config:
            env_vars += container_config['env']

        return env_vars



from datetime import datetime
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.shared.hash import ignore_keys, merge_dict
from typing import Dict, List, Any
import mage_ai


def create_dag(
    project_path: str,
    pipeline_uuid: str,
    dag_class,
    python_operator_class,
    dag_settings: Dict[str, Any] = dict(),
    globals_dict: Dict[str, Any] = dict(),
):
    pipeline = Pipeline(pipeline_uuid, repo_path=project_path)

    tasks = []

    def build_execute_block(block):
        # Airflow v1.x will error unless we use *args instead of ds explicitly.
        def _callable(*args, **kwargs):
            mage_ai.run(
                pipeline_uuid,
                project_path=project_path,
                block_uuid=block.uuid,
                analyze_outputs=False,
                update_status=False,
            )

        return _callable

    for uuid, b in pipeline.blocks_by_uuid.items():
        if b.type == 'scratchpad':
            continue
        tasks.append(dict(
            task_id=uuid,
            upstream_task_ids=b.upstream_block_uuids,
            python_callable=build_execute_block(b),
        ))

    def initialize_tasks(dag, tasks):
        operators = {}

        for task_dict in tasks:
            task_operator = python_operator_class(dag=dag, **ignore_keys(task_dict, [
                'upstream_task_ids',
            ]))
            operators[task_operator.task_id] = task_operator

        for task_dict in tasks:
            for task_id in task_dict.get('upstream_task_ids', []):
                dag.set_dependency(task_id, task_dict['task_id'])

        return operators

    if len(tasks) >= 1:
        dag_id = f'mage_pipeline_{pipeline_uuid}'
        dag = dag_class(
            f'mage_pipeline_{pipeline_uuid}',
            **merge_dict(
                dict(
                    start_date=datetime(2022, 7, 14),
                    description=f'Mage pipeline: {pipeline_uuid}.',
                    schedule_interval='@once',
                    catchup=False,
                ),
                dag_settings,
            )
        )
        globals_dict[dag_id] = dag

        initialize_tasks(dag, tasks)


def create_dags(
    project_path: str,
    dag_class,
    python_operator_class,
    blacklist_pipelines: List[str] = [],
    dag_settings: Dict[str, Any] = dict(),
    globals_dict: Dict[str, Any] = dict(),
):
    all_pipeline_uuids = Pipeline.get_all_pipelines(project_path)
    for pipeline_uuid in all_pipeline_uuids:
        if pipeline_uuid not in blacklist_pipelines:
            create_dag(
                project_path,
                pipeline_uuid,
                dag_class,
                python_operator_class,
                dag_settings,
                globals_dict,
            )



from mage_ai.data_preparation.models.constants import PipelineType
from mage_ai.data_preparation.models.pipelines.integration_pipeline import IntegrationPipeline
from mage_ai.orchestration.db.models.schedules import PipelineRun, BlockRun
from sqlalchemy import or_
from typing import Dict, List, Tuple
import json
import re

KEY_DESTINATION = 'destinations'
KEY_SOURCE = 'sources'


def calculate_metrics(pipeline_run: PipelineRun) -> Dict:
    pipeline = IntegrationPipeline.get(pipeline_run.pipeline_uuid)

    if PipelineType.INTEGRATION != pipeline.type:
        return

    stream_ors = []
    for s in pipeline.streams():
        stream = s['tap_stream_id']
        stream_ors += [
            BlockRun.block_uuid.contains(f'{pipeline.data_loader.uuid}:{stream}'),
            BlockRun.block_uuid.contains(f'{pipeline.data_exporter.uuid}:{stream}'),
        ]
    all_block_runs = BlockRun.query.filter(
        BlockRun.pipeline_run_id == pipeline_run.id,
        or_(*stream_ors),
    ).all()

    block_runs_by_stream = {}
    for br in all_block_runs:
        block_uuid = br.block_uuid
        parts = block_uuid.split(':')
        stream = parts[1]
        if stream not in block_runs_by_stream:
            block_runs_by_stream[stream] = []
        block_runs_by_stream[stream].append(br)

    for s in pipeline.streams():
        stream = s['tap_stream_id']

        destinations = []
        sources = []

        block_runs = block_runs_by_stream.get(stream, [])
        for br in block_runs:
            logs_arr = br.logs['content'].split('\n')

            if f'{pipeline.data_loader.uuid}:{stream}' in br.block_uuid:
                sources.append(logs_arr)
            elif f'{pipeline.data_exporter.uuid}:{stream}' in br.block_uuid:
                destinations.append(logs_arr)

        block_runs_by_stream[stream] = dict(
            destinations=destinations,
            sources=sources,
        )

    shared_metric_keys = [
        'block_tags',
        'error',
        'errors',
        'message',
    ]

    block_metrics_by_stream = get_metrics(block_runs_by_stream, [
        (KEY_SOURCE, shared_metric_keys + [
            'record',
            'records',
        ]),
        (KEY_DESTINATION, shared_metric_keys + [
            'record',
            'records',
            'records_affected',
            'records_inserted',
            'records_updated',
            'state',
        ]),
    ])

    pipeline_logs_by_stream = {}
    pipeline_logs = pipeline_run.logs['content'].split('\n')
    for pipeline_log in pipeline_logs:
        tags = parse_line(pipeline_log)
        stream = tags.get('stream')
        if not stream:
            continue

        if stream not in pipeline_logs_by_stream:
            pipeline_logs_by_stream[stream] = []

        pipeline_logs_by_stream[stream].append(pipeline_log)

    pipeline_metrics_by_stream = {}
    for s in pipeline.streams():
        stream = s['tap_stream_id']
        logs = pipeline_logs_by_stream.get(stream, [])

        pipeline_metrics_by_stream[stream] = get_metrics(dict(pipeline=dict(pipeline=[logs])), [
            ('pipeline', shared_metric_keys + [
                'bookmarks',
                'number_of_batches',
                'record_counts',
            ]),
        ])['pipeline']['pipeline']

    pipeline_run.update(metrics=dict(
        blocks=block_metrics_by_stream,
        destination=pipeline.destination_uuid,
        pipeline=pipeline_metrics_by_stream,
        source=pipeline.source_uuid,
    ))

    return pipeline_run.metrics


def parse_line(line: str) -> Dict:
    tags = {}
    text = re.sub(r'^[\d]{4}-[\d]{2}-[\d]{2}T[\d]{2}:[\d]{2}:[\d]{2}', '', line).strip()

    try:
        data1 = json.loads(text)
        tags = data1.get('tags', {})
        message = data1.get('message', '')
        try:
            data2 = json.loads(message)
            tags.update(data2.get('tags', {}))
        except json.JSONDecodeError:
            tags.update(data1)
            if 'error_stacktrace' in data1:
                tags['error'] = data1['error_stacktrace']
            if 'error' in data1:
                tags['errors'] = data1['error']
    except json.JSONDecodeError:
        pass

    return tags


def get_metrics(logs_by_uuid: Dict, key_and_key_metrics: List[Tuple[str, List[str]]]) -> Dict:
    metrics = {}

    for uuid in logs_by_uuid.keys():
        metrics[uuid] = {}

        for key, key_metrics in key_and_key_metrics:
            metrics[uuid][key] = {}

            logs_for_uuid = logs_by_uuid[uuid][key]
            for logs in logs_for_uuid:
                temp_metrics = {}

                for idx, l in enumerate(logs):
                    tags = parse_line(l)

                    for key_metric in key_metrics:
                        if key_metric in tags:
                            if key_metric not in temp_metrics or key != KEY_DESTINATION:
                                temp_metrics[key_metric] = [tags[key_metric]]
                            else:
                                temp_metrics[key_metric].append(tags[key_metric])

                for key_metric, value_list in temp_metrics.items():
                    if key_metric not in metrics[uuid][key]:
                        metrics[uuid][key][key_metric] = 0

                    for value in value_list:
                        if type(value) is int:
                            metrics[uuid][key][key_metric] += value
                        else:
                            metrics[uuid][key][key_metric] = value

    return metrics



from dataclasses import dataclass, field
from enum import Enum
from typing import List

from mage_ai.services.email.config import EmailConfig
from mage_ai.services.google_chat.config import GoogleChatConfig
from mage_ai.services.opsgenie.config import OpsgenieConfig
from mage_ai.services.slack.config import SlackConfig
from mage_ai.services.teams.config import TeamsConfig
from mage_ai.shared.config import BaseConfig


class AlertOn(str, Enum):
    PIPELINE_RUN_FAILURE = 'trigger_failure'
    PIPELINE_RUN_SUCCESS = 'trigger_success'
    PIPELINE_RUN_PASSED_SLA = 'trigger_passed_sla'


DEFAULT_ALERT_ON = [
    AlertOn.PIPELINE_RUN_FAILURE,
    AlertOn.PIPELINE_RUN_PASSED_SLA,
]


@dataclass
class MessageTemplate(BaseConfig):
    title: str = None
    summary: str = None
    details: str = None


@dataclass
class MessageTemplates(BaseConfig):
    success: MessageTemplate = None
    failure: MessageTemplate = None
    passed_sla: MessageTemplate = None


@dataclass
class NotificationConfig(BaseConfig):
    alert_on: List[AlertOn] = field(default_factory=lambda: DEFAULT_ALERT_ON)
    email_config: EmailConfig = None
    google_chat_config: GoogleChatConfig = None
    opsgenie_config: OpsgenieConfig = None
    slack_config: SlackConfig = None
    teams_config: TeamsConfig = None
    message_templates: MessageTemplates = None



import os


class SecretsManager:
    def __init__(self):
        from aws_secretsmanager_caching import SecretCache, SecretCacheConfig
        from botocore.config import Config
        import boto3

        region_name = os.getenv('AWS_REGION_NAME', 'us-west-2')
        config = Config(region_name=region_name)
        client = boto3.client('secretsmanager', config=config)
        cache_config = SecretCacheConfig()
        self.cache = SecretCache(config=cache_config, client=client)


secrets_manager = SecretsManager()


def get_secret(secret_id: str, cached=True) -> str:
    if cached:
        return secrets_manager.cache.get_secret_string(secret_id)
    else:
        return get_secret_force(secret_id)


def get_secret_force(secret_id: str) -> str:
    from botocore.config import Config
    import boto3

    region_name = os.getenv('AWS_REGION_NAME', 'us-west-2')
    config = Config(region_name=region_name)
    client = boto3.client('secretsmanager', config=config)

    return client.get_secret_value(
        SecretId=secret_id,
    ).get('SecretString')



from tornado import gen, iostream

from mage_ai.api.utils import authenticate_client_and_token
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.data_preparation.models.variable import VariableType
from mage_ai.orchestration.db.models.oauth import Oauth2Application
from mage_ai.orchestration.db.models.schedules import PipelineRun
from mage_ai.server.api.base import BaseHandler
from mage_ai.settings import REQUIRE_USER_AUTHENTICATION


class ApiDownloadHandler(BaseHandler):
    async def get(self, pipeline_uuid, block_uuid, **kwargs):
        self.set_header('Content-Type', 'text/csv')
        self.set_header('Content-Disposition', 'attachment; filename=' + f'{block_uuid}.csv')

        api_key = self.get_argument('api_key', None, True)
        token = self.get_argument('token', None, True)
        if REQUIRE_USER_AUTHENTICATION:
            authenticated = False
            if api_key and token:
                oauth_client = Oauth2Application.query.filter(
                    Oauth2Application.client_id == api_key,
                ).first()
                if oauth_client:
                    oauth_token, valid = authenticate_client_and_token(oauth_client.id, token)
                    authenticated = valid and \
                        oauth_token and \
                        oauth_token.user
            if not authenticated:
                raise Exception('Unauthorized access to download block output.')

        pipeline = Pipeline.get(pipeline_uuid)
        block = pipeline.get_block(block_uuid)
        pipeline_run_id = self.get_argument('pipeline_run_id', None)
        execution_partition = None
        if pipeline_run_id is not None:
            pipeline_run = PipelineRun.query.get(pipeline_run_id)
            execution_partition = pipeline_run.execution_partition

        if block is None:
            raise Exception(f'Block {block_uuid} does not exist in pipeline {pipeline_uuid}')

        tables = block.get_outputs(
            execution_partition=execution_partition,
            include_print_outputs=False,
            csv_lines_only=True,
            sample=False,
            variable_type=VariableType.DATAFRAME,
        )
        for data in tables:
            table = data.get('table', [])
            line_count = len(table)
            for line in range(0, line_count):
                is_last_line = line == line_count - 1
                try:
                    csv_line = table[line] if is_last_line else table[line] + '\n'
                    self.write(csv_line.encode('UTF-8'))
                except iostream.StreamClosedError:
                    break
                if line % 5000 == 0 or is_last_line:
                    await self.flush()
                    # Sleep for a nanosecond so other handlers can run and avoid blocking
                    await gen.sleep(0.000000001)



from abc import ABC, abstractmethod
from typing import Callable


class Queue(ABC):
    @abstractmethod
    def clean_up_jobs(self):
        pass

    @abstractmethod
    def enqueue(self, job_id: str, target: Callable, *args, **kwargs):
        pass

    @abstractmethod
    def has_job(self, job_id: str):
        pass

    @abstractmethod
    def kill_job(self, job_id: str):
        pass

    def _print(self, msg):
        print(f'[{self.__class__.__name__}] {msg}')



COOKIE_OAUTH_TOKEN = 'oauth_token'
ENDPOINTS_BYPASS_OAUTH_CHECK = []
HEADER_API_KEY = 'X-API-KEY'
HEADER_OAUTH_TOKEN = 'OAUTH-TOKEN'
OAUTH_CLIENT_ID_KEY = 'OAUTH_CLIENT_ID'
URL_PARAMETER_API_KEY = 'api_key'



from typing import Dict

from mage_ai.data_preparation.executors.block_executor import BlockExecutor
from mage_ai.services.aws.ecs import ecs
from mage_ai.shared.hash import merge_dict


class EcsBlockExecutor(BlockExecutor):
    def __init__(self, pipeline, block_uuid: str, execution_partition: str = None):
        super().__init__(pipeline, block_uuid, execution_partition=execution_partition)
        self.executor_config = self.pipeline.repo_config.ecs_config or dict()
        if self.block.executor_config is not None:
            self.executor_config = merge_dict(self.executor_config, self.block.executor_config)

    def execute(
        self,
        block_run_id: int = None,
        global_vars: Dict = None,
        **kwargs,
    ) -> None:
        cmd = self._run_commands(block_run_id, global_vars, **kwargs)
        ecs.run_task(' '.join(cmd), ecs_config=self.executor_config)



"""Add repo path to pipeline schedule

Revision ID: e7beb59b44f9
Revises: 83a74eca6ea1
Create Date: 2023-05-30 19:25:57.578309

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'e7beb59b44f9'
down_revision = '83a74eca6ea1'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('pipeline_schedule', schema=None) as batch_op:
        batch_op.add_column(sa.Column('repo_path', sa.String(length=255), nullable=True))

    with op.batch_alter_table('role', schema=None) as batch_op:
        batch_op.create_index(batch_op.f('ix_role_name'), ['name'], unique=True)

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('role', schema=None) as batch_op:
        batch_op.drop_index(batch_op.f('ix_role_name'))

    with op.batch_alter_table('pipeline_schedule', schema=None) as batch_op:
        batch_op.drop_column('repo_path')

    # ### end Alembic commands ###



class AssociatedUserPresenter():
    def user(self, **kwargs):
        user = self.resource.user
        roles_new = [role.to_dict() for role in user.roles_new]
        if user:
            return {
                'avatar': user.avatar,
                'first_name': user.first_name,
                'id': user.id,
                'last_name': user.last_name,
                'owner': user.owner,
                'project_access': user.project_access,
                'roles': user.roles,
                'roles_display': user.roles_display,
                'roles_new': roles_new,
                'username': user.username,
            }
        else:
            return None



import setuptools


def readme():
    with open('README.md', encoding='utf8') as f:
        README = f.read()
    return README


requirements = []
with open('requirements.txt') as f:
    for line in f.read().splitlines():
        if line.startswith('# extras'):
            break
        requirements.append(line)

setuptools.setup(
    name='mage-ai',
    # NOTE: when you change this, change the value of VERSION in the following file:
    # mage_ai/server/constants.py
    version='0.8.93',
    author='Mage',
    author_email='eng@mage.ai',
    description='Mage is a tool for building and deploying data pipelines.',
    long_description=readme(),
    long_description_content_type='text/markdown',
    url='https://github.com/mage-ai/mage-ai',
    packages=setuptools.find_packages('.'),
    include_package_data=True,
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ],
    install_requires=requirements,
    python_requires='>=3.6',
    entry_points={
        'console_scripts': [
            'mage=mage_ai.cli.main:app',
        ],
    },
    extras_require={
        'azure': [
            'azure-eventhub==5.10.1',
            'azure-identity==1.11.0',
            'azure-keyvault-secrets==4.6.0',
            'azure-keyvault-certificates==4.6.0',
            'azure-storage-blob==12.14.1',
        ],
        'bigquery': [
            'google-cloud-bigquery~=3.0',
            'db-dtypes==1.0.5',
        ],
        'clickhouse': [
            'clickhouse-connect==0.5.20',
        ],
        'dbt': [
            'dbt-bigquery==1.4.0',
            'dbt-core==1.4.0',
            'dbt-postgres==1.4.0',
            'dbt-redshift==1.4.0',
            'dbt-snowflake==1.4.0',
            'dbt-spark==1.4.0',
            'dbt-sqlserver==1.3.1',
            'dbt-trino==1.4.0',
        ],
        'google-cloud-storage': ['google-cloud-storage==2.5.0'],
        'hdf5': ['tables==3.7.0'],
        'mysql': [
            'mysql-connector-python==8.0.31',
        ],
        'postgres': [
            'psycopg2==2.9.3',
            'psycopg2-binary==2.9.3',
            'sshtunnel==0.4.0',
        ],
        'redshift': [
            'boto3==1.26.60',
            'redshift-connector==2.0.909',
        ],
        's3': [
            'boto3==1.26.60',
            'botocore==1.29.60',
        ],
        'snowflake': [
            'snowflake-connector-python==2.7.9',
        ],
        'spark': [
            'boto3==1.26.60',
            'botocore==1.29.60',
        ],
        'streaming': [
            'confluent-avro',
            'kafka-python==2.0.2',
            'opensearch-py==2.0.0',
            'pika==1.3.1',
            'pymongo==4.3.3',
            'requests_aws4auth==1.1.2',
        ],
        'all': [
            'aws-secretsmanager-caching==1.1.1.5',
            'azure-eventhub==5.10.1',
            'azure-identity==1.11.0',
            'azure-keyvault-secrets==4.6.0',
            'azure-keyvault-certificates==4.6.0',
            'azure-storage-blob==12.14.1',
            'boto3==1.26.60',
            'botocore==1.29.60',
            'clickhouse-connect==0.5.20',
            'confluent-avro',
            'db-dtypes==1.0.5',
            'dbt-bigquery==1.4.0',
            'dbt-core==1.4.0',
            'dbt-postgres==1.4.0',
            'dbt-redshift==1.4.0',
            'dbt-snowflake==1.4.0',
            'dbt-spark==1.4.0',
            'dbt-sqlserver==1.3.1',
            'dbt-trino==1.4.0',
            'google-api-core==2.11.0',
            'google-api-python-client==2.70.0',
            'google-cloud-bigquery~=3.0',
            'google-cloud-iam==2.9.0',
            'google-cloud-pubsub==2.16.0',
            'google-cloud-run==0.5.0',
            'google-cloud-storage==2.5.0',
            'great_expectations==0.15.50',
            'kafka-python==2.0.2',
            'kubernetes==25.3.0',
            'ldap3==2.9.1',
            'opensearch-py==2.0.0',
            'pika==1.3.1',
            'psycopg2==2.9.3',
            'psycopg2-binary==2.9.3',
            'pydruid==0.6.5',
            'pymongo==4.3.3',
            'pyodbc==4.0.35',
            'redshift-connector==2.0.909',
            'requests_aws4auth==1.1.2',
            'snowflake-connector-python==2.7.9',
            'sshtunnel==0.4.0',
        ],
    },
)



from mage_ai.shared.hash import ignore_keys_with_blank_values
from typing import Dict
import json
import requests


class HttpClient:
    BASE_URL = None

    def __init__(self):
        self.headers = {
            'Content-Type': 'application/json',
        }

    def make_request(
        self,
        url_path: str,
        method: str = 'GET',
        params: Dict = dict(),
        payload: Dict = dict()
    ):
        url = self.BASE_URL + url_path
        response = None
        if method == 'GET':
            response = requests.get(
                url,
                headers=self.headers,
                params=ignore_keys_with_blank_values(params),
            )
        elif method == 'POST':
            response = requests.post(
                url,
                data=json.dumps(ignore_keys_with_blank_values(payload)),
                headers=self.headers,
            )
        if response is not None:
            response.raise_for_status()
            if response.text:
                return response.json()
        return response



from mage_ai.api.oauth_scope import OauthScope
from mage_ai.api.operations import constants
from mage_ai.api.policies.BasePolicy import BasePolicy
from mage_ai.api.presenters.BlockRunPresenter import BlockRunPresenter
from mage_ai.orchestration.db.models.oauth import Permission


class BlockRunPolicy(BasePolicy):
    @property
    def entity(self):
        query = self.options.get('query', {})
        pipeline_uuid = query.get('pipeline_uuid', [None])
        if pipeline_uuid:
            pipeline_uuid = pipeline_uuid[0]
        if pipeline_uuid:
            return Permission.Entity.PIPELINE, pipeline_uuid

        parent_model = self.options.get('parent_model')
        if parent_model:
            return Permission.Entity.PIPELINE, parent_model.pipeline_uuid

        return super().entity


BlockRunPolicy.allow_actions([
    constants.LIST,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_viewer_role())

BlockRunPolicy.allow_read(BlockRunPresenter.default_attributes + [], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.LIST,
], condition=lambda policy: policy.has_at_least_viewer_role())

BlockRunPolicy.allow_query([
    'pipeline_run_id',
    'pipeline_uuid',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_viewer_role())



from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.server.api.base import BaseHandler


class ApiPipelineBlockAnalysisHandler(BaseHandler):
    def get(self, pipeline_uuid, block_uuid):
        pipeline = Pipeline.get(pipeline_uuid)
        block = pipeline.get_block(block_uuid)
        if block is None:
            raise Exception(f'Block {block_uuid} does not exist in pipeline {pipeline_uuid}')
        analyses = block.get_analyses()
        self.write(dict(analyses=analyses))



from mage_ai.errors.base import MageBaseException
from typing import Dict


class DoesNotExistError(MageBaseException):
    pass


class ValidationError(MageBaseException):
    def __init__(self, error: str, metadata: Dict):
        self.error = error
        self.metadata = metadata

    def to_dict(self):
        return dict(
            error=self.error,
            metadata=self.metadata,
        )



from mage_ai.api.oauth_scope import OauthScope
from mage_ai.api.operations import constants
from mage_ai.api.policies.BasePolicy import BasePolicy
from mage_ai.api.presenters.PipelinePresenter import PipelinePresenter
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.orchestration.db.models.oauth import Permission


class PipelinePolicy(BasePolicy):
    @property
    def entity(self):
        if self.resource and self.resource.model:
            return Permission.Entity.PIPELINE, self.resource.model.uuid

        return Permission.Entity.PROJECT, get_repo_path()


PipelinePolicy.allow_actions([
    constants.DETAIL,
    constants.LIST,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_viewer_role())

PipelinePolicy.allow_actions([
    constants.CREATE,
    constants.DELETE,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_editor_role_and_pipeline_edit_access())

PipelinePolicy.allow_actions([
    constants.UPDATE,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_editor_role())

PipelinePolicy.allow_read(PipelinePresenter.default_attributes + [
    'callbacks',
    'conditionals',
    'extensions',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.DETAIL,
], condition=lambda policy: policy.has_at_least_viewer_role())

PipelinePolicy.allow_read(PipelinePresenter.default_attributes + [
    'callbacks',
    'conditionals',
    'extensions',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.CREATE,
    constants.DELETE,
], condition=lambda policy: policy.has_at_least_editor_role_and_pipeline_edit_access())

PipelinePolicy.allow_read(PipelinePresenter.default_attributes + [
    'callbacks',
    'conditionals',
    'extensions',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.UPDATE,
], condition=lambda policy: policy.has_at_least_editor_role())

PipelinePolicy.allow_read(PipelinePresenter.default_attributes + [
    'callbacks',
    'conditionals',
    'extensions',
    'schedules',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.LIST,
], condition=lambda policy: policy.has_at_least_viewer_role())

PipelinePolicy.allow_write([
    'callbacks',
    'conditionals',
    'clone_pipeline_uuid',
    'extensions',
    'name',
    'type',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.CREATE,
], condition=lambda policy: policy.has_at_least_editor_role_and_pipeline_edit_access())

PipelinePolicy.allow_write([
    'add_upstream_for_block_uuid',
    'callbacks',
    'conditionals',
    'extensions',
    'schedules',
] + PipelinePresenter.default_attributes, scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.UPDATE,
], condition=lambda policy: policy.has_at_least_editor_role_and_pipeline_edit_access())

PipelinePolicy.allow_write([
    'pipeline_runs',
    'status',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.UPDATE,
], condition=lambda policy: policy.has_at_least_editor_role())

PipelinePolicy.allow_query([
    'includes_block_metadata',
    'includes_content',
    'includes_extensions',
    'includes_outputs',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.DETAIL,
], condition=lambda policy: policy.has_at_least_viewer_role())

PipelinePolicy.allow_query([
    'status[]',
    'type[]',
    'include_schedules',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.LIST,
], condition=lambda policy: policy.has_at_least_viewer_role())

PipelinePolicy.allow_query([
    'update_content',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.UPDATE,
], condition=lambda policy: policy.has_at_least_editor_role_and_pipeline_edit_access())



from dataclasses import dataclass

from mage_ai.shared.config import BaseConfig


@dataclass
class RetryConfig(BaseConfig):
    retries: int = 0
    delay: int = 5
    max_delay: int = 60
    exponential_backoff: bool = True



from botocore.exceptions import ClientError
from dataclasses import dataclass
from mage_ai.data_preparation.logging.logger_manager import LoggerManager
from mage_ai.data_preparation.repo_manager import RepoConfig
from mage_ai.services.aws.s3 import s3
from mage_ai.shared.config import BaseConfig


@dataclass
class S3Config(BaseConfig):
    bucket: str
    prefix: str


class S3LoggerManager(LoggerManager):
    def __init__(
        self,
        repo_config: RepoConfig = None,
        **kwargs,
    ):
        super().__init__(repo_config=repo_config, **kwargs)
        self.s3_config = S3Config.load(config=self.logging_config.destination_config)
        self.s3_client = s3.Client(bucket=self.s3_config.bucket)

    def output_logs_to_destination(self):
        key = self.get_log_filepath()
        self.s3_client.upload(key, self.stream.getvalue())

    def create_log_filepath_dir(self, path):
        pass

    def get_log_filepath_prefix(self):
        return '{}/{}/{}/{}/{}'.format(
            self.s3_config.prefix,
            self.repo_config.repo_name,
            'pipelines',
            self.pipeline_uuid,
            self.partition,
        )

    def get_logs(self):
        s3_object_key = self.get_log_filepath()
        try:
            s3_object = self.s3_client.read(s3_object_key).decode('UTF-8')
        except ClientError:
            s3_object = ''

        return dict(
            name=s3_object_key.split('/')[-1],
            path=s3_object_key,
            content=s3_object,
        )

    async def get_logs_async(self):
        """
        TODO: Implement this method
        """
        return self.get_logs()



from mage_ai.orchestration.notification.config import NotificationConfig
from mage_ai.tests.base_test import TestCase
from mage_ai.tests.orchestration.notification.constants import (
    EMAIL_NOTIFICATION_CONFIG,
    SLACK_NOTIFICATION_CONFIG,
    TEAMS_NOTIFICATION_CONFIG,
    GOOGLE_CHAT_NOTIFICATION_CONFIG,
    OPSGENIE_NOTIFICATION_CONFIG
)


class NotificationConfigTests(TestCase):

    def test_load_config(self):
        notification_config_empty = dict()
        notification_config_email = EMAIL_NOTIFICATION_CONFIG
        notification_config_slack = SLACK_NOTIFICATION_CONFIG
        notification_config_teams = TEAMS_NOTIFICATION_CONFIG
        notification_config_google_chat = GOOGLE_CHAT_NOTIFICATION_CONFIG
        notification_config_opsgenie = OPSGENIE_NOTIFICATION_CONFIG
        config1 = NotificationConfig.load(config=notification_config_empty)
        config2 = NotificationConfig.load(config=notification_config_email)
        config3 = NotificationConfig.load(config=notification_config_slack)
        config4 = NotificationConfig.load(config=notification_config_teams)
        config5 = NotificationConfig.load(config=notification_config_google_chat)
        config6 = NotificationConfig.load(config=notification_config_opsgenie)

        self.assertIsNone(config1.email_config)
        self.assertIsNone(config1.slack_config)
        self.assertIsNone(config1.teams_config)
        self.assertIsNone(config1.google_chat_config)

        self.assertEqual(config2.email_config.smtp_host, 'test_host')
        self.assertEqual(config2.email_config.smtp_mail_from, 'test_from@abc.com')
        self.assertEqual(config2.email_config.smtp_user, 'test_user')
        self.assertEqual(config2.email_config.smtp_password, 'test_password')
        self.assertEqual(config2.email_config.smtp_starttls, True)
        self.assertEqual(config2.email_config.smtp_ssl, False)
        self.assertEqual(config2.email_config.smtp_port, 587)
        self.assertEqual(config2.email_config.to_emails, ['test_to@xyz.com'])
        self.assertIsNone(config2.slack_config)

        self.assertIsNone(config3.email_config)
        self.assertEqual(config3.slack_config.webhook_url, 'test_webhook_url')

        self.assertEqual(config4.teams_config.webhook_url, 'test_webhook_url')

        self.assertEqual(config5.google_chat_config.webhook_url, 'test_webhook_url')

        self.assertEqual(config6.opsgenie_config.url, 'test_url')
        self.assertEqual(config6.opsgenie_config.api_key, 'test_api_key')



from mage_ai.api.presenters.BasePresenter import BasePresenter


class BlockTemplatePresenter(BasePresenter):
    default_attributes = [
        'block_type',
        'description',
        'groups',
        'language',
        'name',
        'path',
    ]



from mage_ai.streaming.sinks.dummy import DummySink
from mage_ai.tests.base_test import TestCase
from unittest.mock import patch


class DummyTests(TestCase):
    def test_init(self):
        with patch.object(DummySink, 'init_client') as mock_init_client:
            DummySink(dict(
                connector_type='dummy',
                print_msg=True,
            ))
            mock_init_client.assert_called_once()

    def test_init_invalid_config(self):
        with patch.object(DummySink, 'init_client') as mock_init_client:
            with self.assertRaises(Exception) as context:
                DummySink(dict(
                    connector_type='dummy',
                    print_msg=True,
                    test_key='test_value',
                ))
            print(context.exception)
            self.assertTrue(
                '__init__() got an unexpected keyword argument \'test_key\''
                in str(context.exception),
            )
            self.assertEqual(mock_init_client.call_count, 0)



from mage_ai.api.oauth_scope import OauthScope
from mage_ai.api.operations import constants
from mage_ai.api.policies.BasePolicy import BasePolicy
from mage_ai.api.presenters.WidgetPresenter import WidgetPresenter
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.orchestration.db.models.oauth import Permission


class WidgetPolicy(BasePolicy):
    @property
    def entity(self):
        parent_model = self.options.get('parent_model')
        if parent_model:
            return Permission.Entity.PIPELINE, parent_model.uuid

        return Permission.Entity.PROJECT, get_repo_path()


WidgetPolicy.allow_actions([
    constants.LIST,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_viewer_role())

WidgetPolicy.allow_actions([
    constants.CREATE,
    constants.DELETE,
    constants.UPDATE,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_editor_role_and_pipeline_edit_access())

WidgetPolicy.allow_read(WidgetPresenter.default_attributes + [], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.LIST,
], condition=lambda policy: policy.has_at_least_viewer_role())

WidgetPolicy.allow_read(WidgetPresenter.default_attributes + [
    'retry_config',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.CREATE,
    constants.DELETE,
    constants.UPDATE,
], condition=lambda policy: policy.has_at_least_editor_role_and_pipeline_edit_access())

WidgetPolicy.allow_write([
    'all_upstream_blocks_executed',
    'color',
    'config',
    'configuration',
    'content',
    'downstream_blocks',
    'executor_config',
    'executor_type',
    'has_callback',
    'language',
    'name',
    'priority',
    'retry_config',
    'status',
    'type',
    'upstream_blocks',
    'uuid',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.CREATE,
    constants.UPDATE,
], condition=lambda policy: policy.has_at_least_editor_role_and_pipeline_edit_access())



from mage_ai.api.oauth_scope import OauthScope
from mage_ai.api.operations import constants
from mage_ai.api.policies.BasePolicy import BasePolicy
from mage_ai.api.presenters.LogPresenter import LogPresenter
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.orchestration.db.models.oauth import Permission
from mage_ai.orchestration.db.models.schedules import BlockRun


class LogPolicy(BasePolicy):
    @property
    def entity(self):
        parent_model = self.options.get('parent_model')
        if parent_model:
            if type(parent_model) is BlockRun:
                return Permission.Entity.PIPELINE, parent_model.pipeline_run.pipeline_uuid
            elif issubclass(parent_model.__class__, Pipeline):
                return Permission.Entity.PIPELINE, parent_model.uuid

        return Permission.Entity.PROJECT, get_repo_path()


LogPolicy.allow_actions([
    constants.LIST,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_viewer_role())

LogPolicy.allow_read(LogPresenter.default_attributes + [
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.LIST,
], condition=lambda policy: policy.has_at_least_viewer_role())

LogPolicy.allow_query([
    'block_run_id',
    'block_run_id[]',
    'block_type',
    'block_type[]',
    'block_uuid',
    'block_uuid[]',
    'end_timestamp',
    'level[]',
    'pipeline_run_id',
    'pipeline_run_id[]',
    'pipeline_schedule_id',
    'pipeline_schedule_id[]',
    'start_timestamp',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.LIST,
], condition=lambda policy: policy.has_at_least_viewer_role())
